"""A 7x8 14-segment (plus decimal point) display gadget."""

import numpy as np

from ..colors import BLACK, Color
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .text import Text

__all__ = ["DigitalDisplay", "Point", "Size"]

DIM_GREEN = Color.from_hex("062b0f")
BRIGHT_GREEN = Color.from_hex("33e860")

_CHAR_TO_SEGMENTS = {
    " ": (),
    "0": ("a", "b", "c", "d", "e", "f"),
    "1": ("b", "c"),
    "2": ("a", "b", "g2", "g1", "e", "d"),
    "3": ("a", "b", "g2", "c", "d"),
    "4": ("f", "b", "g1", "g2", "c"),
    "5": ("a", "f", "g1", "m", "d"),
    "6": ("a", "f", "g1", "g2", "e", "c", "d"),
    "7": ("a", "b", "c"),
    "8": ("a", "f", "b", "g1", "g2", "e", "c", "d"),
    "9": ("a", "f", "b", "g1", "g2", "c", "d"),
    "a": ("e", "g1", "l", "d"),
    "b": ("f", "g1", "e", "m", "d"),
    "c": ("g1", "g2", "e", "d"),
    "d": ("k", "g2", "b", "c", "d"),
    "e": ("g1", "e", "k", "d"),
    "f": ("g1", "j", "g2", "l"),
    "g": ("j", "b", "g2", "c", "d"),
    "h": ("f", "g1", "e", "l"),
    "i": ("l",),
    "j": ("e", "k", "i"),
    "k": ("i", "l", "j", "m"),
    "l": ("f", "e"),
    "m": ("e", "g1", "l", "g2", "c"),
    "n": ("e", "g1", "l"),
    "o": ("g1", "g2", "e", "c", "d"),
    "p": ("f", "h", "g1", "e"),
    "q": ("j", "b", "g2", "c"),
    "r": ("e", "g1"),
    "s": ("g2", "m", "d"),
    "t": ("f", "g1", "e", "d"),
    "u": ("e", "d", "c"),
    "v": ("e", "k"),
    "w": ("e", "k", "m", "c"),
    "x": ("h", "j", "k", "m"),
    "y": ("i", "g2", "b", "c", "d"),
    "z": ("g1", "k", "d"),
    "A": ("a", "f", "b", "g1", "g2", "e", "c"),
    "B": ("a", "i", "b", "g2", "l", "c", "d"),
    "C": ("a", "f", "e", "d"),
    "D": ("a", "i", "b", "l", "c", "d"),
    "E": ("a", "f", "g1", "e", "d"),
    "F": ("a", "f", "g1", "e"),
    "G": ("a", "f", "g2", "e", "c", "d"),
    "H": ("f", "b", "g1", "g2", "e", "c"),
    "I": ("a", "i", "l", "d"),
    "J": ("e", "d", "c", "b"),
    "K": ("f", "g1", "j", "e", "m"),
    "L": ("f", "e", "d"),
    "M": ("f", "h", "j", "b", "e", "c"),
    "N": ("f", "h", "b", "e", "m", "c"),
    "O": ("a", "f", "b", "e", "c", "d"),
    "P": ("a", "f", "b", "g1", "g2", "e"),
    "Q": ("a", "f", "b", "e", "m", "c", "d"),
    "R": ("a", "f", "b", "g1", "g2", "e", "m"),
    "S": ("a", "f", "g1", "g2", "c", "d"),
    "T": ("a", "i", "l"),
    "U": ("f", "b", "e", "c", "d"),
    "V": ("f", "e", "k", "j"),
    "W": ("f", "b", "e", "k", "m", "c"),
    "X": ("h", "j", "k", "m"),
    "Y": ("f", "b", "g1", "g2", "c", "d"),
    "Z": ("a", "j", "k", "d"),
    "!": ("b", "c", "dp"),
    '"': ("i", "b"),
    "#": ("g1", "i", "g2", "b", "l", "c", "d"),
    "$": ("a", "f", "i", "g1", "g2", "l", "c", "d"),
    "%": ("f", "h", "j", "g1", "g2", "k", "m", "c"),
    "&": ("a", "h", "i", "g1", "e", "m", "d", "c"),
    "'": ("i",),
    "(": ("j", "m"),
    ")": ("h", "k"),
    "*": ("h", "i", "j", "g1", "g2", "k", "l", "m"),
    "+": ("i", "g1", "g2", "l"),
    ",": ("k",),
    "-": ("g1", "g2"),
    ".": ("dp",),
    "/": ("k", "j"),
    ":": ("i", "l"),
    ";": ("i", "k"),
    "<": ("g1", "j", "m"),
    "=": ("g1", "g2", "d"),
    ">": ("h", "k", "g2"),
    "?": ("a", "b", "g2", "l", "dp"),
    "@": ("i", "g2", "b", "a", "f", "e", "d"),
    "[": ("a", "f", "e", "d"),
    "\\": ("h", "m"),
    "]": ("a", "b", "c", "d"),
    "^": ("k", "m"),
    "_": ("d",),
    "`": ("h",),
    "{": ("a", "h", "g1", "k", "d"),
    "|": ("i", "l"),
    "}": ("a", "j", "g2", "m", "d"),
    "~": ("g1", "j", "g2", "k"),
}


class _Segment:
    def __init__(self, slice_):
        self.slice = slice_

    def __get__(self, owner, instance):
        segment = instance._display.canvas["fg_color"][self.slice]
        return (segment == instance.on_color).all()

    def __set__(self, instance, value):
        fg = instance._display.canvas["fg_color"]
        fg[self.slice] = instance.on_color if value else instance.off_color


class DigitalDisplay(Gadget):
    r"""
    A 7x8 14-segment (plus decimal point) display gadget.

    Segments are labeled according to the following diagram::

            a
          ━━━━━
        f┃  ┃i ┃b     h \   / j
         ┃  ┃  ┃         \ /
                     g1 ━━ ━━ g2
        e┃  ┃l ┃c        / \
         ┃  ┃  ┃      k /   \ m
          ━━━━━ ● dp
            d

    Parameters
    ----------
    off_color : Color, default: DIM_GREEN
        Color of off segments.
    on_color : Color, default: BRIGHT_GREEN
        Color of on segments.
    bg_color : Color, default: BLACK
        Background color of gadget.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    off_color : Color
        Color of off segments.
    on_color : Color
        Color of on segments.
    bg_color : Color
        Background color of gadget.
    alpha : float
        Transparency of gadget.
    a : bool
        Whether the `a` segment is on.
    b : bool
        Whether the `b` segment is on.
    c : bool
        Whether the `c` segment is on.
    d : bool
        Whether the `d` segment is on.
    e : bool
        Whether the `e` segment is on.
    f : bool
        Whether the `f` segment is on.
    g1 : bool
        Whether the `g1` segment is on.
    g2 : bool
        Whether the `g2` segment is on.
    h : bool
        Whether the `h` segment is on.
    i : bool
        Whether the `i` segment is on.
    j : bool
        Whether the `j` segment is on.
    k : bool
        Whether the `k` segment is on.
    l : bool
        Whether the `l` segment is on.
    m : bool
        Whether the `m` segment is on.
    dp : bool
        Whether the `dp` segment is on.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    show_char(char)
        Display an ascii character.
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    a = _Segment(np.s_[0, 1:6])
    b = _Segment(np.s_[1:3, 6])
    c = _Segment(np.s_[4:6, 6])
    d = _Segment(np.s_[6, 1:6])
    e = _Segment(np.s_[4:6, 0])
    f = _Segment(np.s_[1:3, 0])
    g1 = _Segment(np.s_[3, 1:3])
    g2 = _Segment(np.s_[3, 4:6])
    h = _Segment(np.s_[[1, 2], [1, 2]])
    i = _Segment(np.s_[1:3, 3])
    j = _Segment(np.s_[[1, 2], [5, 4]])
    k = _Segment(np.s_[[4, 5], [2, 1]])
    l = _Segment(np.s_[4:6, 3])  # noqa
    m = _Segment(np.s_[[4, 5], [4, 5]])
    dp = _Segment(np.s_[6, 7:8])

    def __init__(
        self,
        *,
        off_color: Color = DIM_GREEN,
        on_color: Color = BRIGHT_GREEN,
        bg_color: Color = BLACK,
        alpha: float = 1.0,
        size: Size = Size(7, 8),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._display = Text(
            pos_hint={"y_hint": 0.5, "x_hint": 0.5}, is_transparent=is_transparent
        )
        self._display.set_text(
            " ━━━━━\n"
            "┃\\ ┃ /┃\n"
            "┃ \\┃/ ┃\n"
            " ━━ ━━\n"
            "┃ /┃\\ ┃\n"
            "┃/ ┃ \\┃\n"
            " ━━━━━ ●"
        )
        self._display.canvas["fg_color"] = off_color
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.add_gadget(self._display)
        self._off_color = off_color
        self._on_color = on_color
        self.bg_color = bg_color
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._display.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._display.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._display.is_transparent = self.is_transparent

    @property
    def off_color(self) -> Color:
        """Color of off segments."""
        return self._off_color

    @off_color.setter
    def off_color(self, off_color: Color):
        fg_color = self._display.canvas["fg_color"]
        mask = np.all(fg_color == self.off_color, axis=-1)
        fg_color[mask] = off_color
        self._off_color = off_color

    @property
    def on_color(self) -> Color:
        """Color of on segments."""
        return self._on_color

    @on_color.setter
    def on_color(self, on_color: Color):
        fg_color = self._display.canvas["fg_color"]
        mask = np.all(fg_color == self.on_color, axis=-1)
        fg_color[mask] = on_color
        self._on_color = on_color

    @property
    def bg_color(self) -> Color:
        """Background color of gadget."""
        return self._bg_color

    @bg_color.setter
    def bg_color(self, bg_color: Color):
        self._display.canvas["bg_color"] = bg_color
        self._bg_color = bg_color

    def show_char(self, char: str):
        """
        Show an ascii character in the digital display.

        Parameters
        ----------
        char : str
            The character to display.
        """
        if char not in _CHAR_TO_SEGMENTS:
            raise ValueError(f"{char} is not an ascii character")

        self._display.canvas["fg_color"] = self.off_color

        for segment in _CHAR_TO_SEGMENTS[char]:
            setattr(self, segment, True)



"""A module for constant motion along a bezier curve."""

import asyncio
from bisect import bisect
from collections.abc import Callable
from dataclasses import dataclass
from itertools import accumulate
from math import comb
from time import perf_counter
from typing import Protocol

import numpy as np
from numpy.typing import NDArray

from .basic import Point, clamp
from .easings import EASINGS, Easing

__all__ = ["BezierCurve", "Easing", "move_along_path"]


@dataclass
class BezierCurve:
    """
    A Bezier curve.

    Parameters
    ----------
    control_points : NDArray[np.float32]
        Array of control points of Bezier curve with shape `(N, 2)`.
    arc_length_approximation : int, default: 50
        Number of evaluations for arc length approximation.

    Attributes
    ----------
    arc_length : float
        Approximate length of Bezier curve.
    arc_length_approximation : int
        Number of evaluations for arc length approximation.
    arc_lengths : NDArray[np.float32]
        Approximate arc lengths along Bezier curve.
    coef : NDArray[np.float32]
        Binomial coefficients of Bezier curve.
    control_points : NDArray[np.float32]
        Array of control points of Bezier curve with shape `(N, 2)`.
    degree : int
        Degree of Bezier curve.

    Methods
    -------
    evaluate(t)
        Evaluate the Bezier curve at `t` (0 <= t <= 1).
    arc_length_proportion(p)
        Evaluate the Bezier curve at a proportion of its total arc length.
    """

    control_points: NDArray[np.float32]
    """Array of control points of Bezier curve with shape `(N, 2)`."""
    arc_length_approximation: int = 50
    """Number of evaluations for arc length approximation."""

    def __post_init__(self):
        if self.degree == -1:
            raise ValueError("There must be at least one control point.")

        self.coef: NDArray[np.float32] = np.array(
            [comb(self.degree, i) for i in range(self.degree + 1)], dtype=float
        )
        """Binomial coefficients of Bezier curve."""

        evaluated = self.evaluate(np.linspace(0, 1, self.arc_length_approximation))
        norms = np.linalg.norm(evaluated[1:] - evaluated[:-1], axis=-1)
        self.arc_lengths: NDArray[np.float32] = np.append(0, norms.cumsum())
        """Approximate arc lengths along Bezier curve."""

    @property
    def degree(self) -> int:
        """Degree of Bezier curve."""
        return len(self.control_points) - 1

    @property
    def arc_length(self) -> float:
        """Approximate length of Bezier curve."""
        return self.arc_lengths[-1]

    def evaluate(self, t: float | NDArray[np.float32]) -> NDArray[np.float32]:
        """Evaluate the Bezier curve at `t` (0 <= t <= 1)."""
        t = np.asarray(t)
        terms = np.logspace(0, self.degree, num=self.degree + 1, base=t).T
        terms *= np.logspace(self.degree, 0, num=self.degree + 1, base=1 - t).T
        terms *= self.coef
        return terms @ self.control_points

    def arc_length_proportion(self, p: float) -> NDArray[np.float32]:
        """Evaluate the Bezier curve at a proportion of its total arc length."""
        target_length = self.arc_length * p
        n = self.arc_length_approximation
        i = clamp(bisect(self.arc_lengths, target_length) - 1, 0, n - 1)

        previous_length = self.arc_lengths[i]
        if previous_length == target_length:
            return self.evaluate(i / n)

        target_dif = target_length - previous_length
        if i < n - 1:
            arc_dif = self.arc_lengths[i + 1] - previous_length
        else:
            arc_dif = previous_length - self.arc_lengths[i - 1]

        t = (i + target_dif / arc_dif) / n
        return self.evaluate(t)


class HasPos(Protocol):
    """An object with a position."""

    pos: Point


async def move_along_path(
    has_pos: HasPos,
    path: list[BezierCurve],
    *,
    speed: float = 1.0,
    easing: Easing = "linear",
    on_start: Callable[[], None] | None = None,
    on_progress: Callable[[float], None] | None = None,
    on_complete: Callable[[], None] | None = None,
):
    """
    Move `has_pos` along a path of Bezier curves at some speed (in cells per second).

    Parameters
    ----------
    has_pos : HasPos
        Object to be moved along path.
    path : list[BezierCurve]
        A path made up of Bezier curves.
    speed : float, default: 1.0
        Speed of movement in approximately cells per second.
    on_start : Callable[[], None] | None, default: None
        Called when motion starts.
    on_progress : Callable[[float], None] | None, default: None
        Called as motion updates with current progress.
    on_complete : Callable[[], None] | None, default: None
        Called when motion completes.
    """
    cumulative_arc_lengths = [
        *accumulate((curve.arc_length for curve in path), initial=0)
    ]
    total_arc_length = cumulative_arc_lengths[-1]
    easing_function = EASINGS[easing]
    has_pos.pos = path[0].evaluate(0.0)
    last_time = perf_counter()
    distance_traveled = 0.0

    if on_start is not None:
        on_start()

    while True:
        await asyncio.sleep(0)

        current_time = perf_counter()
        elapsed = current_time - last_time
        last_time = current_time
        distance_traveled += speed * elapsed
        if distance_traveled >= total_arc_length:
            has_pos.pos = path[-1].evaluate(1.0)
            break

        p = easing_function(distance_traveled / total_arc_length)
        eased_distance = p * total_arc_length

        i = clamp(bisect(cumulative_arc_lengths, eased_distance) - 1, 0, len(path) - 1)
        distance_on_curve = eased_distance - cumulative_arc_lengths[i]
        curve_p = distance_on_curve / path[i].arc_length
        has_pos.pos = path[i].arc_length_proportion(curve_p)

        if on_progress is not None:
            on_progress(p)

    if on_complete is not None:
        on_complete()



"""A video gadget."""

import asyncio
import atexit
import time
from pathlib import Path
from platform import uname

import cv2

from ..colors import ABLACK, AColor
from ..logging import get_logger
from ..texture_tools import resize_texture
from .graphics import (
    Blitter,
    Graphics,
    Interpolation,
    Point,
    PosHint,
    Size,
    SizeHint,
    scale_geometry,
)

__all__ = ["Video", "Interpolation", "Point", "Size"]

_IS_WSL: bool = uname().system == "Linux" and uname().release.endswith("Microsoft")

logger = get_logger(__name__)


class Video(Graphics):
    r"""
    A video gadget.

    Parameters
    ----------
    source : pathlib.Path | str | int
        A path to video, URL to video stream, or video capturing device (by index).
        Trying to open a video capturing device on WSL will issue a warning.
    loop : bool, default: True
        Whether to restart video after last frame.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    source : Path | str | int
        A path, URL, or capturing device (by index) of the video.
    loop : bool
        Whether to restart video after last frame.
    is_device : bool
        Whether video is from a video capturing device.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    play()
        Play the video. Returns a task.
    pause()
        Pause the video.
    seek()
        Seek to certain time (in seconds) in the video.
    stop()
        Stop the video.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        source: Path | str | int,
        loop: bool = True,
        default_color: AColor = ABLACK,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._current_frame = None
        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._resource = None
        self._video_task = None
        self.source = source
        self.loop = loop

    @property
    def source(self) -> Path | str | int:
        """A path, URL, or capturing device (by index) of the video."""
        return self._source

    @source.setter
    def source(self, source: Path | str | int):
        self.pause()
        self._release_resource()
        if isinstance(source, Path) and not source.exists():
            raise FileNotFoundError(str(source))
        self._source = source
        self._load_resource()

    @property
    def is_device(self):
        """Return true if source is a video capturing device."""
        return isinstance(self._source, int)

    def _load_resource(self):
        source = self.source

        if _IS_WSL and self.is_device:
            # Because WSL doesn't support most USB devices (yet?), and trying to open
            # one with cv2 will pollute the terminal with cv2 errors, we don't attempt
            # to open a device in this case.
            logger.info("Device not available on WSL.")
            self._resource = None
            return

        if isinstance(source, Path):
            source = str(source.absolute())

        self._resource = cv2.VideoCapture(source)
        atexit.register(self._resource.release)

    def _release_resource(self):
        if self._resource is not None:
            self._resource.release()
            atexit.unregister(self._resource.release)
            self._resource = None
            self._current_frame = None
            self.clear()

    def _time_delta(self) -> float:
        return time.perf_counter() - self._resource.get(cv2.CAP_PROP_POS_MSEC) / 1000

    def _display_current_frame(self):
        h, w = self.size
        if self._current_frame is None or h == 0 or w == 0:
            return

        resize_texture(
            self._current_frame,
            scale_geometry(self._blitter, self._size),
            self._interpolation,
            out=self.texture,
        )

    async def _play_video(self):
        if self._resource is None:
            return

        self._start_time = self._time_delta()

        while True:
            if not self._resource.grab():
                if self.loop:
                    self.seek(0)
                else:
                    self._current_frame = None
                    self.clear()
                    return

            if self.is_device:
                seconds_ahead = 0
            elif (seconds_ahead := self._start_time - self._time_delta()) < 0:
                continue

            try:
                await asyncio.sleep(seconds_ahead)
            except asyncio.CancelledError:
                return

            _, frame = self._resource.retrieve()
            self._current_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)
            self._display_current_frame()

    def on_size(self):
        """Resize current frame on resize."""
        super().on_size()
        self._display_current_frame()

    def on_remove(self):
        """Pause video and release resource on remove."""
        super().on_remove()
        self.pause()
        self._release_resource()

    def pause(self):
        """Pause video."""
        if self._video_task is not None:
            self._video_task.cancel()

    def play(self) -> asyncio.Task:
        """
        Play video.

        Returns
        -------
        asyncio.Task
            The task that plays the video.
        """
        self.pause()

        if self._resource is None:
            self._load_resource()

        self._video_task = asyncio.create_task(self._play_video())
        return self._video_task

    def seek(self, time: float):
        """If supported, seek to certain time (in seconds) in the video."""
        if self._resource is not None and not self.is_device:
            self._resource.set(cv2.CAP_PROP_POS_MSEC, time * 1000)
            self._resource.grab()
            self._start_time = self._time_delta()

    def stop(self):
        """Stop video."""
        self.pause()
        self.seek(0)
        self._current_frame = None
        self.clear()



"""Horizontal and vertical stack layout gadgets."""

import math

from ..geometry import Point, Size
from .gadget import Gadget, PosHint, SizeHint

__all__ = ["VStackLayout", "HStackLayout", "Point", "Size"]


class _StackLayoutBase(Gadget):
    def __init__(
        self,
        *,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._proportions = ()

    @property
    def proportions(self) -> tuple[float, ...]:
        """Size of children as proportions of the stack layout size."""
        return self._proportions

    @proportions.setter
    def proportions(self, proportions: tuple[float, ...]):
        if not self.children:
            self._proportions = ()
            return

        total_proportion = sum(proportions)
        if len(proportions) != len(self.children) or total_proportion == 0:
            self._proportions = (1 / len(self.children),) * len(self.children)
        elif not math.isclose(total_proportion, 1):
            if total_proportion == 0:
                self._proportions = (1 / len(self.children),) * len(self.children)
            else:
                factor = 1 / total_proportion
                self._proportions = tuple(
                    proportion * factor for proportion in proportions
                )
        self._reposition_children()

    def on_size(self):
        """Resize children on resize."""
        self.proportions = self.proportions

    def add_gadget(self, gadget: Gadget):
        """Remove hints from gadget and resize children when child is added."""
        gadget.size_hint = {}
        gadget.pos_hint = {}
        super().add_gadget(gadget)
        self.proportions = self.proportions

    def remove_gadget(self, gadget: Gadget):
        """Resize children when child is removed."""
        super().remove_gadget(gadget)
        self.proportions = self.proportions


class VStackLayout(_StackLayoutBase):
    r"""
    A vertical stack layout gadget.

    A vertical stack layout positions its children vertically and resizes them
    to use all of its available space. By default, each child will take an
    (almost) equal amount of space, but this can be adjusted by setting
    :attr:`proportions`. The nth proportion adjusts the height of the nth child
    gadget. Extra unfilled rows or columns are divided among the children.

    Parameters
    ----------
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    proportions : tuple[float, ...]
        Height of children as proportions of stack layout height.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.

    Notes
    -----
    Stack layouts remove size and pos hints from their children.
    """

    def _reposition_children(self):
        for proportion, child in zip(self._proportions, self.children):
            child.size = int(proportion * self.height), self.width

        total_height = sum(child.height for child in self.children)
        offset_all, offset_few = divmod(self.height - total_height, len(self.children))
        for i, child in enumerate(reversed(self.children)):
            child.height += offset_all + i < offset_few

        y = 0
        for child in self.children:
            child.pos = y, 0
            y += child.height


class HStackLayout(_StackLayoutBase):
    r"""
    A horizontal stack layout gadget.

    A horizontal stack layout positions its children horizontally and resizes
    them to use all of its available space. By default, each child will take an
    (almost) equal amount of space, but this can be adjusted by setting
    :attr:`proportions`. The nth proportion adjusts the width of the nth child
    gadget. Extra unfilled rows or columns are divided among the children.

    Parameters
    ----------
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    proportions : tuple[float, ...]
        Width of children as proportions of stack layout width.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.

    Notes
    -----
    Stack layouts remove size and pos hints from their children.
    """

    def _reposition_children(self):
        for proportion, child in zip(self._proportions, self.children):
            child.size = self.height, int(proportion * self.width)

        total_width = sum(child.width for child in self.children)
        offset_all, offset_few = divmod(self.width - total_width, len(self.children))
        for i, child in enumerate(reversed(self.children)):
            child.width += offset_all + i < offset_few

        x = 0
        for child in self.children:
            child.pos = 0, x
            x += child.width



"""
A text particle field.

A particle field specializes in rendering many single "pixel" children from an array of
particle positions and an array of particle cells.
"""

from typing import Any

import numpy as np
from numpy.typing import NDArray

from .._rendering import text_field_render
from ..geometry import clamp
from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint, new_cell

__all__ = ["TextParticleField", "Point", "Size"]


class TextParticleField(Gadget):
    r"""
    A text particle field.

    A particle field specializes in rendering many single "pixel" children from an array
    of particle positions and an array of particle cells.

    Parameters
    ----------
    particle_coords : NDArray[np.float64] | None, default: None
        An array of particle positions with shape `(N, 2)`.
    particle_cells : NDArray[Cell] | None, default: None
        An array of Cells of particles with shape `(N,)`.
    particle_properties : dict[str, Any] | None, default: None
        Additional particle properties.
    alpha : float, default: 0.0
        Transparency of particles.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    nparticles : int
        Number of particles in particle field.
    particle_coords : NDArray[np.float64]
        An array of particle positions with shape `(N, 2)`.
    particle_cells : NDArray[Cell]
        An array of Cells of particles with shape `(N,)`.
    particle_properties : dict[str, Any]
        Additional particle properties.
    alpha : float
        Transparency of particles.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    particles_from_cells(cells)
        Set particle positions and cells from non-whitespace characters of a Cell array.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        particle_coords: NDArray[np.float64] | None = None,
        particle_cells: NDArray[Cell] | None = None,
        particle_properties: dict[str, Any] = None,
        alpha: float = 0.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.particle_coords: NDArray[np.float64]
        """An array of particle positions with shape `(N, 2)`."""
        if particle_coords is None:
            self.particle_coords = np.zeros((0, 2), dtype=np.float64)
        else:
            self.particle_coords = np.ascontiguousarray(
                particle_coords, dtype=np.float64
            )

        self.particle_cells: NDArray[Cell]
        """An array of Cells of particles with shape `(N,)`"""
        if particle_cells is None:
            self.particle_cells = np.full(len(self.particle_coords), new_cell())
        else:
            self.particle_cells = np.ascontiguousarray(particle_cells, dtype=Cell)

        self.particle_properties: dict[str, Any]
        """Additional particle properties."""
        if particle_properties is None:
            self.particle_properties = {}
        else:
            self.particle_properties = particle_properties

        self.alpha = alpha
        """Transparency of gadget."""

    @property
    def alpha(self) -> float:
        """Transparency of particles."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._alpha = clamp(alpha, 0.0, 1.0)

    @property
    def nparticles(self) -> int:
        """Number of particles in particle field."""
        return len(self.particle_coords)

    def particles_from_cells(self, cells: NDArray[Cell]) -> None:
        """
        Set particle positions and cells from non-whitespace characters of a Cell array.

        Parameters
        ----------
        cells : NDArray[Cell]
            A Cell numpy array (such as a :class:`Text` gadget's canvas).
        """
        positions = np.argwhere(np.isin(cells["ord"], (32, 0x2800), invert=True))
        pys, pxs = positions.T
        self.particle_coords = np.ascontiguousarray(positions.astype(np.float64))
        self.particle_cells = np.ascontiguousarray(cells[pys, pxs])

    def _render(
        self,
        cells: NDArray[Cell],
        graphics: NDArray[np.uint8],
        kind: NDArray[np.uint8],
    ):
        """Render visible region of gadget."""
        text_field_render(
            cells,
            graphics,
            kind,
            self.absolute_pos,
            self._is_transparent,
            self.particle_coords,
            self.particle_cells,
            self._alpha,
            self._region,
        )



"""A gadget with a background color that is composited if transparent."""

import numpy as np
from numpy.typing import NDArray

from .._rendering import pane_render
from ..colors import BLACK, Color
from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint, bindable, clamp

__all__ = ["Pane", "Point", "Size"]


class Pane(Gadget):
    r"""
    A gadget with a background color that is composited if transparent.

    Parameters
    ----------
    bg_color : Color, default: BLACK
        Background color of gadget.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    bg_color : Color
        Background color of gadget.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        bg_color: Color = BLACK,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.bg_color: Color = bg_color
        """Background color of gadget."""
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)

    def _render(
        self, cells: NDArray[Cell], graphics: NDArray[np.uint8], kind: NDArray[np.uint8]
    ) -> None:
        """Render visible region of gadget."""
        pane_render(
            cells,
            graphics,
            kind,
            self._is_transparent,
            self._region,
            self.bg_color,
            self.alpha,
        )



"""
Text effects are coroutines that create some effect on a text gadget.

This module is an approximate recreation of some effects from terminaltexteffects.

References
----------
https://github.com/ChrisBuilds/terminaltexteffects

Warnings
--------
Modifying the text size while effect is running will break the effect.
"""

from .beams import beams_effect
from .black_hole import black_hole_effect
from .ring import ring_effect
from .spotlights import spotlights_effect

__all__ = ["beams_effect", "black_hole_effect", "ring_effect", "spotlights_effect"]



"""Markdown emoji codes."""

# Source: https://github.com/markdown-templates/markdown-emojis

EMOJIS = {
    "100": "💯",
    "1234": "🔢",
    "interrobang": "⁉️",
    "tm": "™️",
    "information_source": "ℹ️",
    "left_right_arrow": "↔️",
    "arrow_up_down": "↕️",
    "arrow_upper_left": "↖️",
    "arrow_upper_right": "↗️",
    "arrow_lower_right": "↘️",
    "arrow_lower_left": "↙️",
    "keyboard": "⌨️",
    "sunny": "☀️",
    "cloud": "☁️",
    "umbrella_with_rain_drops": "☂️",
    "snowman": "☃️",
    "comet": "☄️",
    "ballot_box_with_check": "☑️",
    "umbrella": "☔️",
    "coffee": "☕️",
    "shamrock": "☘️",
    "skull_and_crossbones": "☠️",
    "radioactive_sign": "☢️",
    "biohazard_sign": "☣️",
    "orthodox_cross": "☦️",
    "wheel_of_dharma": "☸️",
    "white_frowning_face": "☹️",
    "female_sign": "♀",
    "male_sign": "♂",
    "aries": "♈️",
    "taurus": "♉️",
    "sagittarius": "♐️",
    "capricorn": "♑️",
    "aquarius": "♒️",
    "pisces": "♓️",
    "spades": "♠️",
    "clubs": "♣️",
    "hearts": "♥️",
    "diamonds": "♦️",
    "hotsprings": "♨️",
    "hammer_and_pick": "⚒",
    "anchor": "⚓️",
    "crossed_swords": "⚔️",
    "staff_of_aesculapius": "⚕",
    "scales": "⚖️",
    "alembic": "⚗️",
    "gear": "⚙️",
    "scissors": "✂️",
    "white_check_mark": "✅",
    "airplane": "✈️",
    "email": "✉️",
    "envelope": "✉️",
    "black_nib": "✒️",
    "heavy_check_mark": "✔️",
    "heavy_multiplication_x": "✖️",
    "star_of_david": "✡️",
    "sparkles": "✨",
    "eight_spoked_asterisk": "✳️",
    "eight_pointed_black_star": "✴️",
    "snowflake": "❄️",
    "sparkle": "❇️",
    "question": "❓",
    "grey_question": "❔",
    "grey_exclamation": "❕",
    "exclamation": "❗️",
    "heavy_exclamation_mark": "❗️",
    "heavy_heart_exclamation_mark_ornament": "❣️",
    "heart": "❤️",
    "heavy_plus_sign": "➕",
    "heavy_minus_sign": "➖",
    "heavy_division_sign": "➗",
    "arrow_heading_up": "⤴️",
    "arrow_heading_down": "⤵️",
    "wavy_dash": "〰️",
    "congratulations": "㊗️",
    "secret": "㊙️",
    "copyright": "©️",
    "registered": "®️",
    "bangbang": "‼️",
    "leftwards_arrow_with_hook": "↩️",
    "arrow_right_hook": "↪️",
    "watch": "⌚️",
    "hourglass": "⌛️",
    "eject": "⏏",
    "fast_forward": "⏩",
    "rewind": "⏪",
    "arrow_double_up": "⏫",
    "arrow_double_down": "⏬",
    "black_right_pointing_double_triangle_with_vertical_bar": "⏭",
    "black_left_pointing_double_triangle_with_vertical_bar": "⏮",
    "black_right_pointing_triangle_with_double_vertical_bar": "⏯",
    "alarm_clock": "⏰",
    "stopwatch": "⏱",
    "timer_clock": "⏲",
    "hourglass_flowing_sand": "⏳",
    "double_vertical_bar": "⏸",
    "black_square_for_stop": "⏹",
    "black_circle_for_record": "⏺",
    "m": "Ⓜ️",
    "black_small_square": "▪️",
    "white_small_square": "▫️",
    "arrow_forward": "▶️",
    "arrow_backward": "◀️",
    "white_medium_square": "◻️",
    "black_medium_square": "◼️",
    "white_medium_small_square": "◽️",
    "black_medium_small_square": "◾️",
    "phone": "☎️",
    "telephone": "☎️",
    "point_up": "☝️",
    "star_and_crescent": "☪️",
    "peace_symbol": "☮️",
    "yin_yang": "☯️",
    "relaxed": "☺️",
    "gemini": "♊️",
    "cancer": "♋️",
    "leo": "♌️",
    "virgo": "♍️",
    "libra": "♎️",
    "scorpius": "♏️",
    "recycle": "♻️",
    "wheelchair": "♿️",
    "atom_symbol": "⚛️",
    "fleur_de_lis": "⚜️",
    "warning": "⚠️",
    "zap": "⚡️",
    "white_circle": "⚪️",
    "black_circle": "⚫️",
    "coffin": "⚰️",
    "funeral_urn": "⚱️",
    "soccer": "⚽️",
    "baseball": "⚾️",
    "snowman_without_snow": "⛄️",
    "partly_sunny": "⛅️",
    "thunder_cloud_and_rain": "⛈",
    "ophiuchus": "⛎",
    "pick": "⛏",
    "helmet_with_white_cross": "⛑",
    "chains": "⛓",
    "no_entry": "⛔️",
    "shinto_shrine": "⛩",
    "church": "⛪️",
    "mountain": "⛰",
    "umbrella_on_ground": "⛱",
    "fountain": "⛲️",
    "golf": "⛳️",
    "ferry": "⛴",
    "boat": "⛵️",
    "sailboat": "⛵️",
    "skier": "⛷",
    "ice_skate": "⛸",
    "tent": "⛺️",
    "fuelpump": "⛽️",
    "fist": "✊",
    "hand": "✋",
    "raised_hand": "✋",
    "v": "✌️",
    "writing_hand": "✍️",
    "pencil2": "✏️",
    "latin_cross": "✝️",
    "x": "❌",
    "negative_squared_cross_mark": "❎",
    "arrow_right": "➡️",
    "curly_loop": "➰",
    "loop": "➿",
    "arrow_left": "⬅️",
    "arrow_up": "⬆️",
    "arrow_down": "⬇️",
    "black_large_square": "⬛️",
    "white_large_square": "⬜️",
    "star": "⭐️",
    "o": "⭕️",
    "part_alternation_mark": "〽️",
    "mahjong": "🀄️",
    "black_joker": "🃏",
    "a": "🅰️",
    "b": "🅱️",
    "o2": "🅾️",
    "parking": "🅿️",
    "ab": "🆎",
    "cl": "🆑",
    "cool": "🆒",
    "free": "🆓",
    "id": "🆔",
    "new": "🆕",
    "ng": "🆖",
    "ok": "🆗",
    "sos": "🆘",
    "up": "🆙",
    "vs": "🆚",
    "koko": "🈁",
    "sa": "🈂️",
    "u7121": "🈚️",
    "u6307": "🈯️",
    "u7981": "🈲",
    "u7a7a": "🈳",
    "u5408": "🈴",
    "u6e80": "🈵",
    "u6709": "🈶",
    "u6708": "🈷️",
    "u7533": "🈸",
    "u5272": "🈹",
    "u55b6": "🈺",
    "ideograph_advantage": "🉐",
    "accept": "🉑",
    "cyclone": "🌀",
    "foggy": "🌁",
    "closed_umbrella": "🌂",
    "night_with_stars": "🌃",
    "sunrise_over_mountains": "🌄",
    "sunrise": "🌅",
    "city_sunset": "🌆",
    "city_sunrise": "🌇",
    "rainbow": "🌈",
    "bridge_at_night": "🌉",
    "ocean": "🌊",
    "volcano": "🌋",
    "milky_way": "🌌",
    "earth_africa": "🌍",
    "earth_americas": "🌎",
    "earth_asia": "🌏",
    "globe_with_meridians": "🌐",
    "new_moon": "🌑",
    "waxing_crescent_moon": "🌒",
    "first_quarter_moon": "🌓",
    "moon": "🌔",
    "waxing_gibbous_moon": "🌔",
    "full_moon": "🌕",
    "waning_gibbous_moon": "🌖",
    "last_quarter_moon": "🌗",
    "waning_crescent_moon": "🌘",
    "crescent_moon": "🌙",
    "new_moon_with_face": "🌚",
    "first_quarter_moon_with_face": "🌛",
    "last_quarter_moon_with_face": "🌜",
    "full_moon_with_face": "🌝",
    "sun_with_face": "🌞",
    "star2": "🌟",
    "stars": "🌠",
    "thermometer": "🌡",
    "mostly_sunny": "🌤",
    "sun_small_cloud": "🌤",
    "barely_sunny": "🌥",
    "sun_behind_cloud": "🌥",
    "partly_sunny_rain": "🌦",
    "sun_behind_rain_cloud": "🌦",
    "rain_cloud": "🌧",
    "snow_cloud": "🌨",
    "lightning": "🌩",
    "lightning_cloud": "🌩",
    "tornado": "🌪",
    "tornado_cloud": "🌪",
    "fog": "🌫",
    "wind_blowing_face": "🌬",
    "hotdog": "🌭",
    "taco": "🌮",
    "burrito": "🌯",
    "chestnut": "🌰",
    "seedling": "🌱",
    "evergreen_tree": "🌲",
    "deciduous_tree": "🌳",
    "palm_tree": "🌴",
    "cactus": "🌵",
    "hot_pepper": "🌶",
    "tulip": "🌷",
    "cherry_blossom": "🌸",
    "rose": "🌹",
    "hibiscus": "🌺",
    "sunflower": "🌻",
    "blossom": "🌼",
    "corn": "🌽",
    "ear_of_rice": "🌾",
    "herb": "🌿",
    "four_leaf_clover": "🍀",
    "maple_leaf": "🍁",
    "fallen_leaf": "🍂",
    "leaves": "🍃",
    "mushroom": "🍄",
    "tomato": "🍅",
    "eggplant": "🍆",
    "grapes": "🍇",
    "melon": "🍈",
    "watermelon": "🍉",
    "tangerine": "🍊",
    "lemon": "🍋",
    "banana": "🍌",
    "pineapple": "🍍",
    "apple": "🍎",
    "green_apple": "🍏",
    "pear": "🍐",
    "peach": "🍑",
    "cherries": "🍒",
    "strawberry": "🍓",
    "hamburger": "🍔",
    "pizza": "🍕",
    "meat_on_bone": "🍖",
    "poultry_leg": "🍗",
    "rice_cracker": "🍘",
    "rice_ball": "🍙",
    "rice": "🍚",
    "curry": "🍛",
    "ramen": "🍜",
    "spaghetti": "🍝",
    "bread": "🍞",
    "fries": "🍟",
    "sweet_potato": "🍠",
    "dango": "🍡",
    "oden": "🍢",
    "sushi": "🍣",
    "fried_shrimp": "🍤",
    "fish_cake": "🍥",
    "icecream": "🍦",
    "shaved_ice": "🍧",
    "ice_cream": "🍨",
    "doughnut": "🍩",
    "cookie": "🍪",
    "chocolate_bar": "🍫",
    "candy": "🍬",
    "lollipop": "🍭",
    "custard": "🍮",
    "honey_pot": "🍯",
    "cake": "🍰",
    "bento": "🍱",
    "stew": "🍲",
    "fried_egg": "🍳",
    "cooking": "🍳",
    "fork_and_knife": "🍴",
    "tea": "🍵",
    "sake": "🍶",
    "wine_glass": "🍷",
    "cocktail": "🍸",
    "tropical_drink": "🍹",
    "beer": "🍺",
    "beers": "🍻",
    "baby_bottle": "🍼",
    "knife_fork_plate": "🍽",
    "champagne": "🍾",
    "popcorn": "🍿",
    "ribbon": "🎀",
    "gift": "🎁",
    "birthday": "🎂",
    "jack_o_lantern": "🎃",
    "christmas_tree": "🎄",
    "santa": "🎅",
    "fireworks": "🎆",
    "sparkler": "🎇",
    "balloon": "🎈",
    "tada": "🎉",
    "confetti_ball": "🎊",
    "tanabata_tree": "🎋",
    "crossed_flags": "🎌",
    "bamboo": "🎍",
    "dolls": "🎎",
    "flags": "🎏",
    "wind_chime": "🎐",
    "rice_scene": "🎑",
    "school_satchel": "🎒",
    "mortar_board": "🎓",
    "medal": "🎖",
    "reminder_ribbon": "🎗",
    "studio_microphone": "🎙",
    "level_slider": "🎚",
    "control_knobs": "🎛",
    "film_frames": "🎞",
    "admission_tickets": "🎟",
    "carousel_horse": "🎠",
    "ferris_wheel": "🎡",
    "roller_coaster": "🎢",
    "fishing_pole_and_fish": "🎣",
    "microphone": "🎤",
    "movie_camera": "🎥",
    "cinema": "🎦",
    "headphones": "🎧",
    "art": "🎨",
    "tophat": "🎩",
    "circus_tent": "🎪",
    "ticket": "🎫",
    "clapper": "🎬",
    "performing_arts": "🎭",
    "video_game": "🎮",
    "dart": "🎯",
    "slot_machine": "🎰",
    "8ball": "🎱",
    "game_die": "🎲",
    "bowling": "🎳",
    "flower_playing_cards": "🎴",
    "musical_note": "🎵",
    "notes": "🎶",
    "saxophone": "🎷",
    "guitar": "🎸",
    "musical_keyboard": "🎹",
    "trumpet": "🎺",
    "violin": "🎻",
    "musical_score": "🎼",
    "running_shirt_with_sash": "🎽",
    "tennis": "🎾",
    "ski": "🎿",
    "basketball": "🏀",
    "checkered_flag": "🏁",
    "snowboarder": "🏂",
    "sports_medal": "🏅",
    "trophy": "🏆",
    "horse_racing": "🏇",
    "football": "🏈",
    "rugby_football": "🏉",
    "racing_motorcycle": "🏍",
    "racing_car": "🏎",
    "cricket_bat_and_ball": "🏏",
    "volleyball": "🏐",
    "field_hockey_stick_and_ball": "🏑",
    "ice_hockey_stick_and_puck": "🏒",
    "table_tennis_paddle_and_ball": "🏓",
    "snow_capped_mountain": "🏔",
    "camping": "🏕",
    "beach_with_umbrella": "🏖",
    "building_construction": "🏗",
    "house_buildings": "🏘",
    "cityscape": "🏙",
    "derelict_house_building": "🏚",
    "classical_building": "🏛",
    "desert": "🏜",
    "desert_island": "🏝",
    "national_park": "🏞",
    "stadium": "🏟",
    "house": "🏠",
    "house_with_garden": "🏡",
    "office": "🏢",
    "post_office": "🏣",
    "european_post_office": "🏤",
    "hospital": "🏥",
    "bank": "🏦",
    "atm": "🏧",
    "hotel": "🏨",
    "love_hotel": "🏩",
    "convenience_store": "🏪",
    "school": "🏫",
    "department_store": "🏬",
    "factory": "🏭",
    "izakaya_lantern": "🏮",
    "lantern": "🏮",
    "japanese_castle": "🏯",
    "european_castle": "🏰",
    "waving_white_flag": "🏳️",
    "waving_black_flag": "🏴",
    "rosette": "🏵",
    "label": "🏷",
    "badminton_racquet_and_shuttlecock": "🏸",
    "bow_and_arrow": "🏹",
    "amphora": "🏺",
    "skin-tone-2": "🏻",
    "skin-tone-3": "🏼",
    "skin-tone-4": "🏽",
    "skin-tone-5": "🏾",
    "skin-tone-6": "🏿",
    "rat": "🐀",
    "mouse2": "🐁",
    "ox": "🐂",
    "water_buffalo": "🐃",
    "cow2": "🐄",
    "tiger2": "🐅",
    "leopard": "🐆",
    "rabbit2": "🐇",
    "cat2": "🐈",
    "dragon": "🐉",
    "crocodile": "🐊",
    "whale2": "🐋",
    "snail": "🐌",
    "snake": "🐍",
    "racehorse": "🐎",
    "ram": "🐏",
    "goat": "🐐",
    "sheep": "🐑",
    "monkey": "🐒",
    "rooster": "🐓",
    "chicken": "🐔",
    "dog2": "🐕",
    "pig2": "🐖",
    "boar": "🐗",
    "elephant": "🐘",
    "octopus": "🐙",
    "shell": "🐚",
    "bug": "🐛",
    "ant": "🐜",
    "bee": "🐝",
    "honeybee": "🐝",
    "beetle": "🐞",
    "fish": "🐟",
    "tropical_fish": "🐠",
    "blowfish": "🐡",
    "turtle": "🐢",
    "hatching_chick": "🐣",
    "baby_chick": "🐤",
    "hatched_chick": "🐥",
    "bird": "🐦",
    "penguin": "🐧",
    "koala": "🐨",
    "poodle": "🐩",
    "dromedary_camel": "🐪",
    "camel": "🐫",
    "dolphin": "🐬",
    "flipper": "🐬",
    "mouse": "🐭",
    "cow": "🐮",
    "tiger": "🐯",
    "rabbit": "🐰",
    "cat": "🐱",
    "dragon_face": "🐲",
    "whale": "🐳",
    "horse": "🐴",
    "monkey_face": "🐵",
    "dog": "🐶",
    "pig": "🐷",
    "frog": "🐸",
    "hamster": "🐹",
    "wolf": "🐺",
    "bear": "🐻",
    "panda_face": "🐼",
    "pig_nose": "🐽",
    "feet": "🐾",
    "paw_prints": "🐾",
    "chipmunk": "🐿",
    "eyes": "👀",
    "eye": "👁",
    "ear": "👂",
    "nose": "👃",
    "lips": "👄",
    "tongue": "👅",
    "point_up_2": "👆",
    "point_down": "👇",
    "point_left": "👈",
    "point_right": "👉",
    "facepunch": "👊",
    "punch": "👊",
    "wave": "👋",
    "ok_hand": "👌",
    "+1": "👍",
    "thumbsup": "👍",
    "-1": "👎",
    "thumbsdown": "👎",
    "clap": "👏",
    "open_hands": "👐",
    "crown": "👑",
    "womans_hat": "👒",
    "eyeglasses": "👓",
    "necktie": "👔",
    "shirt": "👕",
    "tshirt": "👕",
    "jeans": "👖",
    "dress": "👗",
    "kimono": "👘",
    "bikini": "👙",
    "womans_clothes": "👚",
    "purse": "👛",
    "handbag": "👜",
    "pouch": "👝",
    "mans_shoe": "👞",
    "shoe": "👞",
    "athletic_shoe": "👟",
    "high_heel": "👠",
    "sandal": "👡",
    "boot": "👢",
    "footprints": "👣",
    "bust_in_silhouette": "👤",
    "busts_in_silhouette": "👥",
    "boy": "👦",
    "girl": "👧",
    "man": "👨",
    "woman": "👩",
    "couple": "👫",
    "man_and_woman_holding_hands": "👫",
    "two_men_holding_hands": "👬",
    "two_women_holding_hands": "👭",
    "bride_with_veil": "👰",
    "man_with_gua_pi_mao": "👲",
    "older_man": "👴",
    "older_woman": "👵",
    "baby": "👶",
    "princess": "👸",
    "japanese_ogre": "👹",
    "japanese_goblin": "👺",
    "ghost": "👻",
    "angel": "👼",
    "alien": "👽",
    "space_invader": "👾",
    "imp": "👿",
    "skull": "💀",
    "dancer": "💃",
    "lipstick": "💄",
    "nail_care": "💅",
    "barber": "💈",
    "syringe": "💉",
    "pill": "💊",
    "kiss": "💋",
    "love_letter": "💌",
    "ring": "💍",
    "gem": "💎",
    "bouquet": "💐",
    "wedding": "💒",
    "heartbeat": "💓",
    "broken_heart": "💔",
    "two_hearts": "💕",
    "sparkling_heart": "💖",
    "heartpulse": "💗",
    "cupid": "💘",
    "blue_heart": "💙",
    "green_heart": "💚",
    "yellow_heart": "💛",
    "purple_heart": "💜",
    "gift_heart": "💝",
    "revolving_hearts": "💞",
    "heart_decoration": "💟",
    "diamond_shape_with_a_dot_inside": "💠",
    "bulb": "💡",
    "anger": "💢",
    "bomb": "💣",
    "zzz": "💤",
    "boom": "💥",
    "collision": "💥",
    "sweat_drops": "💦",
    "droplet": "💧",
    "dash": "💨",
    "hankey": "💩",
    "poop": "💩",
    "shit": "💩",
    "muscle": "💪",
    "dizzy": "💫",
    "speech_balloon": "💬",
    "thought_balloon": "💭",
    "white_flower": "💮",
    "moneybag": "💰",
    "currency_exchange": "💱",
    "heavy_dollar_sign": "💲",
    "credit_card": "💳",
    "yen": "💴",
    "dollar": "💵",
    "euro": "💶",
    "pound": "💷",
    "money_with_wings": "💸",
    "chart": "💹",
    "seat": "💺",
    "computer": "💻",
    "briefcase": "💼",
    "minidisc": "💽",
    "floppy_disk": "💾",
    "cd": "💿",
    "dvd": "📀",
    "file_folder": "📁",
    "open_file_folder": "📂",
    "page_with_curl": "📃",
    "page_facing_up": "📄",
    "date": "📅",
    "calendar": "📆",
    "card_index": "📇",
    "chart_with_upwards_trend": "📈",
    "chart_with_downwards_trend": "📉",
    "bar_chart": "📊",
    "clipboard": "📋",
    "pushpin": "📌",
    "round_pushpin": "📍",
    "paperclip": "📎",
    "straight_ruler": "📏",
    "triangular_ruler": "📐",
    "bookmark_tabs": "📑",
    "ledger": "📒",
    "notebook": "📓",
    "notebook_with_decorative_cover": "📔",
    "closed_book": "📕",
    "book": "📖",
    "open_book": "📖",
    "green_book": "📗",
    "blue_book": "📘",
    "orange_book": "📙",
    "books": "📚",
    "name_badge": "📛",
    "scroll": "📜",
    "memo": "📝",
    "pencil": "📝",
    "telephone_receiver": "📞",
    "pager": "📟",
    "fax": "📠",
    "satellite_antenna": "📡",
    "loudspeaker": "📢",
    "mega": "📣",
    "outbox_tray": "📤",
    "inbox_tray": "📥",
    "package": "📦",
    "e-mail": "📧",
    "incoming_envelope": "📨",
    "envelope_with_arrow": "📩",
    "mailbox_closed": "📪",
    "mailbox": "📫",
    "mailbox_with_mail": "📬",
    "mailbox_with_no_mail": "📭",
    "postbox": "📮",
    "postal_horn": "📯",
    "newspaper": "📰",
    "iphone": "📱",
    "calling": "📲",
    "vibration_mode": "📳",
    "mobile_phone_off": "📴",
    "no_mobile_phones": "📵",
    "signal_strength": "📶",
    "camera": "📷",
    "camera_with_flash": "📸",
    "video_camera": "📹",
    "tv": "📺",
    "radio": "📻",
    "vhs": "📼",
    "film_projector": "📽",
    "prayer_beads": "📿",
    "twisted_rightwards_arrows": "🔀",
    "repeat": "🔁",
    "repeat_one": "🔂",
    "arrows_clockwise": "🔃",
    "arrows_counterclockwise": "🔄",
    "low_brightness": "🔅",
    "high_brightness": "🔆",
    "mute": "🔇",
    "speaker": "🔈",
    "sound": "🔉",
    "loud_sound": "🔊",
    "battery": "🔋",
    "electric_plug": "🔌",
    "mag": "🔍",
    "mag_right": "🔎",
    "lock_with_ink_pen": "🔏",
    "closed_lock_with_key": "🔐",
    "key": "🔑",
    "lock": "🔒",
    "unlock": "🔓",
    "bell": "🔔",
    "no_bell": "🔕",
    "bookmark": "🔖",
    "link": "🔗",
    "radio_button": "🔘",
    "back": "🔙",
    "end": "🔚",
    "on": "🔛",
    "soon": "🔜",
    "top": "🔝",
    "underage": "🔞",
    "keycap_ten": "🔟",
    "capital_abcd": "🔠",
    "abcd": "🔡",
    "symbols": "🔣",
    "abc": "🔤",
    "fire": "🔥",
    "flashlight": "🔦",
    "wrench": "🔧",
    "hammer": "🔨",
    "nut_and_bolt": "🔩",
    "hocho": "🔪",
    "knife": "🔪",
    "gun": "🔫",
    "microscope": "🔬",
    "telescope": "🔭",
    "crystal_ball": "🔮",
    "six_pointed_star": "🔯",
    "beginner": "🔰",
    "trident": "🔱",
    "black_square_button": "🔲",
    "white_square_button": "🔳",
    "red_circle": "🔴",
    "large_blue_circle": "🔵",
    "large_orange_diamond": "🔶",
    "large_blue_diamond": "🔷",
    "small_orange_diamond": "🔸",
    "small_blue_diamond": "🔹",
    "small_red_triangle": "🔺",
    "small_red_triangle_down": "🔻",
    "arrow_up_small": "🔼",
    "arrow_down_small": "🔽",
    "om_symbol": "🕉",
    "dove_of_peace": "🕊",
    "kaaba": "🕋",
    "mosque": "🕌",
    "synagogue": "🕍",
    "menorah_with_nine_branches": "🕎",
    "clock1": "🕐",
    "clock2": "🕑",
    "clock3": "🕒",
    "clock4": "🕓",
    "clock5": "🕔",
    "clock6": "🕕",
    "clock7": "🕖",
    "clock8": "🕗",
    "clock9": "🕘",
    "clock10": "🕙",
    "clock11": "🕚",
    "clock12": "🕛",
    "clock130": "🕜",
    "clock230": "🕝",
    "clock330": "🕞",
    "clock430": "🕟",
    "clock530": "🕠",
    "clock630": "🕡",
    "clock730": "🕢",
    "clock830": "🕣",
    "clock930": "🕤",
    "clock1030": "🕥",
    "clock1130": "🕦",
    "clock1230": "🕧",
    "candle": "🕯",
    "mantelpiece_clock": "🕰",
    "hole": "🕳",
    "man_in_business_suit_levitating": "🕴",
    "dark_sunglasses": "🕶",
    "spider": "🕷",
    "spider_web": "🕸",
    "joystick": "🕹",
    "man_dancing": "🕺",
    "linked_paperclips": "🖇",
    "lower_left_ballpoint_pen": "🖊",
    "lower_left_fountain_pen": "🖋",
    "lower_left_paintbrush": "🖌",
    "lower_left_crayon": "🖍",
    "raised_hand_with_fingers_splayed": "🖐",
    "middle_finger": "🖕",
    "reversed_hand_with_middle_finger_extended": "🖕",
    "spock-hand": "🖖",
    "black_heart": "🖤",
    "desktop_computer": "🖥",
    "printer": "🖨",
    "three_button_mouse": "🖱",
    "trackball": "🖲",
    "frame_with_picture": "🖼",
    "card_index_dividers": "🗂",
    "card_file_box": "🗃",
    "file_cabinet": "🗄",
    "wastebasket": "🗑",
    "spiral_note_pad": "🗒",
    "spiral_calendar_pad": "🗓",
    "compression": "🗜",
    "old_key": "🗝",
    "rolled_up_newspaper": "🗞",
    "dagger_knife": "🗡",
    "speaking_head_in_silhouette": "🗣",
    "left_speech_bubble": "🗨",
    "right_anger_bubble": "🗯",
    "ballot_box_with_ballot": "🗳",
    "world_map": "🗺",
    "mount_fuji": "🗻",
    "tokyo_tower": "🗼",
    "statue_of_liberty": "🗽",
    "japan": "🗾",
    "moyai": "🗿",
    "grinning": "😀",
    "grin": "😁",
    "joy": "😂",
    "smiley": "😃",
    "smile": "😄",
    "sweat_smile": "😅",
    "laughing": "😆",
    "satisfied": "😆",
    "innocent": "😇",
    "smiling_imp": "😈",
    "wink": "😉",
    "blush": "😊",
    "yum": "😋",
    "relieved": "😌",
    "heart_eyes": "😍",
    "sunglasses": "😎",
    "smirk": "😏",
    "neutral_face": "😐",
    "expressionless": "😑",
    "unamused": "😒",
    "sweat": "😓",
    "pensive": "😔",
    "confused": "😕",
    "confounded": "😖",
    "kissing": "😗",
    "kissing_heart": "😘",
    "kissing_smiling_eyes": "😙",
    "kissing_closed_eyes": "😚",
    "stuck_out_tongue": "😛",
    "stuck_out_tongue_winking_eye": "😜",
    "stuck_out_tongue_closed_eyes": "😝",
    "disappointed": "😞",
    "worried": "😟",
    "angry": "😠",
    "rage": "😡",
    "cry": "😢",
    "persevere": "😣",
    "triumph": "😤",
    "disappointed_relieved": "😥",
    "frowning": "😦",
    "anguished": "😧",
    "fearful": "😨",
    "weary": "😩",
    "sleepy": "😪",
    "tired_face": "😫",
    "grimacing": "😬",
    "sob": "😭",
    "open_mouth": "😮",
    "hushed": "😯",
    "cold_sweat": "😰",
    "scream": "😱",
    "astonished": "😲",
    "flushed": "😳",
    "sleeping": "😴",
    "dizzy_face": "😵",
    "no_mouth": "😶",
    "mask": "😷",
    "smile_cat": "😸",
    "joy_cat": "😹",
    "smiley_cat": "😺",
    "heart_eyes_cat": "😻",
    "smirk_cat": "😼",
    "kissing_cat": "😽",
    "pouting_cat": "😾",
    "crying_cat_face": "😿",
    "scream_cat": "🙀",
    "slightly_frowning_face": "🙁",
    "slightly_smiling_face": "🙂",
    "upside_down_face": "🙃",
    "face_with_rolling_eyes": "🙄",
    "see_no_evil": "🙈",
    "hear_no_evil": "🙉",
    "speak_no_evil": "🙊",
    "raised_hands": "🙌",
    "pray": "🙏",
    "rocket": "🚀",
    "helicopter": "🚁",
    "steam_locomotive": "🚂",
    "railway_car": "🚃",
    "bullettrain_side": "🚄",
    "bullettrain_front": "🚅",
    "train2": "🚆",
    "metro": "🚇",
    "light_rail": "🚈",
    "station": "🚉",
    "tram": "🚊",
    "train": "🚋",
    "bus": "🚌",
    "oncoming_bus": "🚍",
    "trolleybus": "🚎",
    "busstop": "🚏",
    "minibus": "🚐",
    "ambulance": "🚑",
    "fire_engine": "🚒",
    "police_car": "🚓",
    "oncoming_police_car": "🚔",
    "taxi": "🚕",
    "oncoming_taxi": "🚖",
    "car": "🚗",
    "red_car": "🚗",
    "oncoming_automobile": "🚘",
    "blue_car": "🚙",
    "truck": "🚚",
    "articulated_lorry": "🚛",
    "tractor": "🚜",
    "monorail": "🚝",
    "mountain_railway": "🚞",
    "suspension_railway": "🚟",
    "mountain_cableway": "🚠",
    "aerial_tramway": "🚡",
    "ship": "🚢",
    "speedboat": "🚤",
    "traffic_light": "🚥",
    "vertical_traffic_light": "🚦",
    "construction": "🚧",
    "rotating_light": "🚨",
    "triangular_flag_on_post": "🚩",
    "door": "🚪",
    "no_entry_sign": "🚫",
    "smoking": "🚬",
    "no_smoking": "🚭",
    "put_litter_in_its_place": "🚮",
    "do_not_litter": "🚯",
    "potable_water": "🚰",
    "non-potable_water": "🚱",
    "bike": "🚲",
    "no_bicycles": "🚳",
    "no_pedestrians": "🚷",
    "children_crossing": "🚸",
    "mens": "🚹",
    "womens": "🚺",
    "restroom": "🚻",
    "baby_symbol": "🚼",
    "toilet": "🚽",
    "wc": "🚾",
    "shower": "🚿",
    "bath": "🛀",
    "bathtub": "🛁",
    "passport_control": "🛂",
    "customs": "🛃",
    "baggage_claim": "🛄",
    "left_luggage": "🛅",
    "couch_and_lamp": "🛋",
    "sleeping_accommodation": "🛌",
    "shopping_bags": "🛍",
    "bellhop_bell": "🛎",
    "bed": "🛏",
    "place_of_worship": "🛐",
    "octagonal_sign": "🛑",
    "shopping_trolley": "🛒",
    "hammer_and_wrench": "🛠",
    "shield": "🛡",
    "oil_drum": "🛢",
    "motorway": "🛣",
    "railway_track": "🛤",
    "motor_boat": "🛥",
    "small_airplane": "🛩",
    "airplane_departure": "🛫",
    "airplane_arriving": "🛬",
    "satellite": "🛰",
    "passenger_ship": "🛳",
    "scooter": "🛴",
    "motor_scooter": "🛵",
    "canoe": "🛶",
    "zipper_mouth_face": "🤐",
    "money_mouth_face": "🤑",
    "face_with_thermometer": "🤒",
    "nerd_face": "🤓",
    "thinking_face": "🤔",
    "face_with_head_bandage": "🤕",
    "robot_face": "🤖",
    "hugging_face": "🤗",
    "the_horns": "🤘",
    "sign_of_the_horns": "🤘",
    "call_me_hand": "🤙",
    "raised_back_of_hand": "🤚",
    "left-facing_fist": "🤛",
    "right-facing_fist": "🤜",
    "handshake": "🤝",
    "hand_with_index_and_middle_fingers_crossed": "🤞",
    "face_with_cowboy_hat": "🤠",
    "clown_face": "🤡",
    "nauseated_face": "🤢",
    "rolling_on_the_floor_laughing": "🤣",
    "drooling_face": "🤤",
    "lying_face": "🤥",
    "face_palm": "🤦",
    "sneezing_face": "🤧",
    "pregnant_woman": "🤰",
    "selfie": "🤳",
    "prince": "🤴",
    "man_in_tuxedo": "🤵",
    "mother_christmas": "🤶",
    "shrug": "🤷",
    "person_doing_cartwheel": "🤸",
    "juggling": "🤹",
    "fencer": "🤺",
    "wrestlers": "🤼",
    "water_polo": "🤽",
    "handball": "🤾",
    "wilted_flower": "🥀",
    "drum_with_drumsticks": "🥁",
    "clinking_glasses": "🥂",
    "tumbler_glass": "🥃",
    "spoon": "🥄",
    "goal_net": "🥅",
    "first_place_medal": "🥇",
    "second_place_medal": "🥈",
    "third_place_medal": "🥉",
    "boxing_glove": "🥊",
    "martial_arts_uniform": "🥋",
    "croissant": "🥐",
    "avocado": "🥑",
    "cucumber": "🥒",
    "bacon": "🥓",
    "potato": "🥔",
    "carrot": "🥕",
    "baguette_bread": "🥖",
    "green_salad": "🥗",
    "shallow_pan_of_food": "🥘",
    "stuffed_flatbread": "🥙",
    "egg": "🥚",
    "glass_of_milk": "🥛",
    "peanuts": "🥜",
    "kiwifruit": "🥝",
    "pancakes": "🥞",
    "crab": "🦀",
    "lion_face": "🦁",
    "scorpion": "🦂",
    "turkey": "🦃",
    "unicorn_face": "🦄",
    "eagle": "🦅",
    "duck": "🦆",
    "bat": "🦇",
    "shark": "🦈",
    "owl": "🦉",
    "fox_face": "🦊",
    "butterfly": "🦋",
    "deer": "🦌",
    "gorilla": "🦍",
    "lizard": "🦎",
    "rhinoceros": "🦏",
    "shrimp": "🦐",
    "squid": "🦑",
    "cheese_wedge": "🧀",
    "hash": "#️⃣",
    "keycap_star": "*️⃣",
    "zero": "0️⃣",
    "one": "1️⃣",
    "two": "2️⃣",
    "three": "3️⃣",
    "four": "4️⃣",
    "five": "5️⃣",
    "six": "6️⃣",
    "seven": "7️⃣",
    "eight": "8️⃣",
    "nine": "9️⃣",
    "flag-ac": "🇦🇨",
    "flag-ad": "🇦🇩",
    "flag-ae": "🇦🇪",
    "flag-af": "🇦🇫",
    "flag-ag": "🇦🇬",
    "flag-ai": "🇦🇮",
    "flag-al": "🇦🇱",
    "flag-am": "🇦🇲",
    "flag-ao": "🇦🇴",
    "flag-aq": "🇦🇶",
    "flag-ar": "🇦🇷",
    "flag-as": "🇦🇸",
    "flag-at": "🇦🇹",
    "flag-au": "🇦🇺",
    "flag-aw": "🇦🇼",
    "flag-ax": "🇦🇽",
    "flag-az": "🇦🇿",
    "flag-ba": "🇧🇦",
    "flag-bb": "🇧🇧",
    "flag-bd": "🇧🇩",
    "flag-be": "🇧🇪",
    "flag-bf": "🇧🇫",
    "flag-bg": "🇧🇬",
    "flag-bh": "🇧🇭",
    "flag-bi": "🇧🇮",
    "flag-bj": "🇧🇯",
    "flag-bl": "🇧🇱",
    "flag-bm": "🇧🇲",
    "flag-bn": "🇧🇳",
    "flag-bo": "🇧🇴",
    "flag-bq": "🇧🇶",
    "flag-br": "🇧🇷",
    "flag-bs": "🇧🇸",
    "flag-bt": "🇧🇹",
    "flag-bv": "🇧🇻",
    "flag-bw": "🇧🇼",
    "flag-by": "🇧🇾",
    "flag-bz": "🇧🇿",
    "flag-ca": "🇨🇦",
    "flag-cc": "🇨🇨",
    "flag-cd": "🇨🇩",
    "flag-cf": "🇨🇫",
    "flag-cg": "🇨🇬",
    "flag-ch": "🇨🇭",
    "flag-ci": "🇨🇮",
    "flag-ck": "🇨🇰",
    "flag-cl": "🇨🇱",
    "flag-cm": "🇨🇲",
    "flag-cn": "🇨🇳",
    "cn": "🇨🇳",
    "flag-co": "🇨🇴",
    "flag-cp": "🇨🇵",
    "flag-cr": "🇨🇷",
    "flag-cu": "🇨🇺",
    "flag-cv": "🇨🇻",
    "flag-cw": "🇨🇼",
    "flag-cx": "🇨🇽",
    "flag-cy": "🇨🇾",
    "flag-cz": "🇨🇿",
    "flag-de": "🇩🇪",
    "de": "🇩🇪",
    "flag-dg": "🇩🇬",
    "flag-dj": "🇩🇯",
    "flag-dk": "🇩🇰",
    "flag-dm": "🇩🇲",
    "flag-do": "🇩🇴",
    "flag-dz": "🇩🇿",
    "flag-ea": "🇪🇦",
    "flag-ec": "🇪🇨",
    "flag-ee": "🇪🇪",
    "flag-eg": "🇪🇬",
    "flag-eh": "🇪🇭",
    "flag-er": "🇪🇷",
    "flag-es": "🇪🇸",
    "es": "🇪🇸",
    "flag-et": "🇪🇹",
    "flag-eu": "🇪🇺",
    "flag-fi": "🇫🇮",
    "flag-fj": "🇫🇯",
    "flag-fk": "🇫🇰",
    "flag-fm": "🇫🇲",
    "flag-fo": "🇫🇴",
    "flag-fr": "🇫🇷",
    "fr": "🇫🇷",
    "flag-ga": "🇬🇦",
    "flag-gb": "🇬🇧",
    "gb": "🇬🇧",
    "uk": "🇬🇧",
    "flag-gd": "🇬🇩",
    "flag-ge": "🇬🇪",
    "flag-gf": "🇬🇫",
    "flag-gg": "🇬🇬",
    "flag-gh": "🇬🇭",
    "flag-gi": "🇬🇮",
    "flag-gl": "🇬🇱",
    "flag-gm": "🇬🇲",
    "flag-gn": "🇬🇳",
    "flag-gp": "🇬🇵",
    "flag-gq": "🇬🇶",
    "flag-gr": "🇬🇷",
    "flag-gs": "🇬🇸",
    "flag-gt": "🇬🇹",
    "flag-gu": "🇬🇺",
    "flag-gw": "🇬🇼",
    "flag-gy": "🇬🇾",
    "flag-hk": "🇭🇰",
    "flag-hm": "🇭🇲",
    "flag-hn": "🇭🇳",
    "flag-hr": "🇭🇷",
    "flag-ht": "🇭🇹",
    "flag-hu": "🇭🇺",
    "flag-ic": "🇮🇨",
    "flag-id": "🇮🇩",
    "flag-ie": "🇮🇪",
    "flag-il": "🇮🇱",
    "flag-im": "🇮🇲",
    "flag-in": "🇮🇳",
    "flag-io": "🇮🇴",
    "flag-iq": "🇮🇶",
    "flag-ir": "🇮🇷",
    "flag-is": "🇮🇸",
    "flag-it": "🇮🇹",
    "it": "🇮🇹",
    "flag-je": "🇯🇪",
    "flag-jm": "🇯🇲",
    "flag-jo": "🇯🇴",
    "flag-jp": "🇯🇵",
    "jp": "🇯🇵",
    "flag-ke": "🇰🇪",
    "flag-kg": "🇰🇬",
    "flag-kh": "🇰🇭",
    "flag-ki": "🇰🇮",
    "flag-km": "🇰🇲",
    "flag-kn": "🇰🇳",
    "flag-kp": "🇰🇵",
    "flag-kr": "🇰🇷",
    "kr": "🇰🇷",
    "flag-kw": "🇰🇼",
    "flag-ky": "🇰🇾",
    "flag-kz": "🇰🇿",
    "flag-la": "🇱🇦",
    "flag-lb": "🇱🇧",
    "flag-lc": "🇱🇨",
    "flag-li": "🇱🇮",
    "flag-lk": "🇱🇰",
    "flag-lr": "🇱🇷",
    "flag-ls": "🇱🇸",
    "flag-lt": "🇱🇹",
    "flag-lu": "🇱🇺",
    "flag-lv": "🇱🇻",
    "flag-ly": "🇱🇾",
    "flag-ma": "🇲🇦",
    "flag-mc": "🇲🇨",
    "flag-md": "🇲🇩",
    "flag-me": "🇲🇪",
    "flag-mf": "🇲🇫",
    "flag-mg": "🇲🇬",
    "flag-mh": "🇲🇭",
    "flag-mk": "🇲🇰",
    "flag-ml": "🇲🇱",
    "flag-mm": "🇲🇲",
    "flag-mn": "🇲🇳",
    "flag-mo": "🇲🇴",
    "flag-mp": "🇲🇵",
    "flag-mq": "🇲🇶",
    "flag-mr": "🇲🇷",
    "flag-ms": "🇲🇸",
    "flag-mt": "🇲🇹",
    "flag-mu": "🇲🇺",
    "flag-mv": "🇲🇻",
    "flag-mw": "🇲🇼",
    "flag-mx": "🇲🇽",
    "flag-my": "🇲🇾",
    "flag-mz": "🇲🇿",
    "flag-na": "🇳🇦",
    "flag-nc": "🇳🇨",
    "flag-ne": "🇳🇪",
    "flag-nf": "🇳🇫",
    "flag-ng": "🇳🇬",
    "flag-ni": "🇳🇮",
    "flag-nl": "🇳🇱",
    "flag-no": "🇳🇴",
    "flag-np": "🇳🇵",
    "flag-nr": "🇳🇷",
    "flag-nu": "🇳🇺",
    "flag-nz": "🇳🇿",
    "flag-om": "🇴🇲",
    "flag-pa": "🇵🇦",
    "flag-pe": "🇵🇪",
    "flag-pf": "🇵🇫",
    "flag-pg": "🇵🇬",
    "flag-ph": "🇵🇭",
    "flag-pk": "🇵🇰",
    "flag-pl": "🇵🇱",
    "flag-pm": "🇵🇲",
    "flag-pn": "🇵🇳",
    "flag-pr": "🇵🇷",
    "flag-ps": "🇵🇸",
    "flag-pt": "🇵🇹",
    "flag-pw": "🇵🇼",
    "flag-py": "🇵🇾",
    "flag-qa": "🇶🇦",
    "flag-re": "🇷🇪",
    "flag-ro": "🇷🇴",
    "flag-rs": "🇷🇸",
    "flag-ru": "🇷🇺",
    "ru": "🇷🇺",
    "flag-rw": "🇷🇼",
    "flag-sa": "🇸🇦",
    "flag-sb": "🇸🇧",
    "flag-sc": "🇸🇨",
    "flag-sd": "🇸🇩",
    "flag-se": "🇸🇪",
    "flag-sg": "🇸🇬",
    "flag-sh": "🇸🇭",
    "flag-si": "🇸🇮",
    "flag-sj": "🇸🇯",
    "flag-sk": "🇸🇰",
    "flag-sl": "🇸🇱",
    "flag-sm": "🇸🇲",
    "flag-sn": "🇸🇳",
    "flag-so": "🇸🇴",
    "flag-sr": "🇸🇷",
    "flag-ss": "🇸🇸",
    "flag-st": "🇸🇹",
    "flag-sv": "🇸🇻",
    "flag-sx": "🇸🇽",
    "flag-sy": "🇸🇾",
    "flag-sz": "🇸🇿",
    "flag-ta": "🇹🇦",
    "flag-tc": "🇹🇨",
    "flag-td": "🇹🇩",
    "flag-tf": "🇹🇫",
    "flag-tg": "🇹🇬",
    "flag-th": "🇹🇭",
    "flag-tj": "🇹🇯",
    "flag-tk": "🇹🇰",
    "flag-tl": "🇹🇱",
    "flag-tm": "🇹🇲",
    "flag-tn": "🇹🇳",
    "flag-to": "🇹🇴",
    "flag-tr": "🇹🇷",
    "flag-tt": "🇹🇹",
    "flag-tv": "🇹🇻",
    "flag-tw": "🇹🇼",
    "flag-tz": "🇹🇿",
    "flag-ua": "🇺🇦",
    "flag-ug": "🇺🇬",
    "flag-um": "🇺🇲",
    "flag-un": "🇺🇳",
    "flag-us": "🇺🇸",
    "us": "🇺🇸",
    "flag-uy": "🇺🇾",
    "flag-uz": "🇺🇿",
    "flag-va": "🇻🇦",
    "flag-vc": "🇻🇨",
    "flag-ve": "🇻🇪",
    "flag-vg": "🇻🇬",
    "flag-vi": "🇻🇮",
    "flag-vn": "🇻🇳",
    "flag-vu": "🇻🇺",
    "flag-wf": "🇼🇫",
    "flag-ws": "🇼🇸",
    "flag-xk": "🇽🇰",
    "flag-ye": "🇾🇪",
    "flag-yt": "🇾🇹",
    "flag-za": "🇿🇦",
    "flag-zm": "🇿🇲",
    "flag-zw": "🇿🇼",
    "male-farmer": "👨‍🌾",
    "male-cook": "👨‍🍳",
    "male-student": "👨‍🎓",
    "male-singer": "👨‍🎤",
    "male-artist": "👨‍🎨",
    "male-teacher": "👨‍🏫",
    "male-factory-worker": "👨‍🏭",
    "man-boy": "👨‍👦",
    "man-girl": "👨‍👧",
    "male-technologist": "👨‍💻",
    "male-office-worker": "👨‍💼",
    "male-mechanic": "👨‍🔧",
    "male-scientist": "👨‍🔬",
    "male-astronaut": "👨‍🚀",
    "male-firefighter": "👨‍🚒",
    "female-farmer": "👩‍🌾",
    "female-cook": "👩‍🍳",
    "female-student": "👩‍🎓",
    "female-singer": "👩‍🎤",
    "female-artist": "👩‍🎨",
    "female-teacher": "👩‍🏫",
    "female-factory-worker": "👩‍🏭",
    "woman-boy": "👩‍👦",
    "woman-girl": "👩‍👧",
    "female-technologist": "👩‍💻",
    "female-office-worker": "👩‍💼",
    "female-mechanic": "👩‍🔧",
    "female-scientist": "👩‍🔬",
    "female-astronaut": "👩‍🚀",
    "female-firefighter": "👩‍🚒",
    "woman-running": "🏃‍♀️",
    "man-running": "🏃‍♂️",
    "runner": "🏃‍♂️",
    "running": "🏃‍♂️",
    "woman-surfing": "🏄‍♀️",
    "man-surfing": "🏄‍♂️",
    "surfer": "🏄‍♂️",
    "woman-swimming": "🏊‍♀️",
    "man-swimming": "🏊‍♂️",
    "swimmer": "🏊‍♂️",
    "woman-lifting-weights": "🏋️‍♀️",
    "man-lifting-weights": "🏋️‍♂️",
    "weight_lifter": "🏋️‍♂️",
    "woman-golfing": "🏌️‍♀️",
    "man-golfing": "🏌️‍♂️",
    "golfer": "🏌️‍♂️",
    "rainbow-flag": "🏳️‍🌈",
    "eye-in-speech-bubble": "👁️‍🗨️",
    "man-boy-boy": "👨‍👦‍👦",
    "man-girl-boy": "👨‍👧‍👦",
    "man-girl-girl": "👨‍👧‍👧",
    "man-man-boy": "👨‍👨‍👦",
    "man-man-boy-boy": "👨‍👨‍👦‍👦",
    "man-man-girl": "👨‍👨‍👧",
    "man-man-girl-boy": "👨‍👨‍👧‍👦",
    "man-man-girl-girl": "👨‍👨‍👧‍👧",
    "man-woman-boy": "👨‍👩‍👦",
    "family": "👨‍👩‍👦",
    "man-woman-boy-boy": "👨‍👩‍👦‍👦",
    "man-woman-girl": "👨‍👩‍👧",
    "man-woman-girl-boy": "👨‍👩‍👧‍👦",
    "man-woman-girl-girl": "👨‍👩‍👧‍👧",
    "male-doctor": "👨‍⚕️",
    "male-judge": "👨‍⚖️",
    "male-pilot": "👨‍✈️",
    "man-heart-man": "👨‍❤️‍👨",
    "man-kiss-man": "👨‍❤️‍💋‍👨",
    "woman-boy-boy": "👩‍👦‍👦",
    "woman-girl-boy": "👩‍👧‍👦",
    "woman-girl-girl": "👩‍👧‍👧",
    "woman-woman-boy": "👩‍👩‍👦",
    "woman-woman-boy-boy": "👩‍👩‍👦‍👦",
    "woman-woman-girl": "👩‍👩‍👧",
    "woman-woman-girl-boy": "👩‍👩‍👧‍👦",
    "woman-woman-girl-girl": "👩‍👩‍👧‍👧",
    "female-doctor": "👩‍⚕️",
    "female-judge": "👩‍⚖️",
    "female-pilot": "👩‍✈️",
    "woman-heart-man": "👩‍❤️‍👨",
    "couple_with_heart": "👩‍❤️‍👨",
    "woman-heart-woman": "👩‍❤️‍👩",
    "woman-kiss-man": "👩‍❤️‍💋‍👨",
    "couplekiss": "👩‍❤️‍💋‍👨",
    "woman-kiss-woman": "👩‍❤️‍💋‍👩",
    "female-police-officer": "👮‍♀️",
    "male-police-officer": "👮‍♂️",
    "cop": "👮‍♂️",
    "woman-with-bunny-ears-partying": "👯‍♀️",
    "dancers": "👯‍♀️",
    "man-with-bunny-ears-partying": "👯‍♂️",
    "blond-haired-woman": "👱‍♀️",
    "blond-haired-man": "👱‍♂️",
    "person_with_blond_hair": "👱‍♂️",
    "woman-wearing-turban": "👳‍♀️",
    "man-wearing-turban": "👳‍♂️",
    "man_with_turban": "👳‍♂️",
    "female-construction-worker": "👷‍♀️",
    "male-construction-worker": "👷‍♂️",
    "construction_worker": "👷‍♂️",
    "woman-tipping-hand": "💁‍♀️",
    "information_desk_person": "💁‍♀️",
    "man-tipping-hand": "💁‍♂️",
    "female-guard": "💂‍♀️",
    "male-guard": "💂‍♂️",
    "guardsman": "💂‍♂️",
    "woman-getting-massage": "💆‍♀️",
    "massage": "💆‍♀️",
    "man-getting-massage": "💆‍♂️",
    "woman-getting-haircut": "💇‍♀️",
    "haircut": "💇‍♀️",
    "man-getting-haircut": "💇‍♂️",
    "female-detective": "🕵️‍♀️",
    "male-detective": "🕵️‍♂️",
    "sleuth_or_spy": "🕵️‍♂️",
    "woman-gesturing-no": "🙅‍♀️",
    "no_good": "🙅‍♀️",
    "man-gesturing-no": "🙅‍♂️",
    "woman-gesturing-ok": "🙆‍♀️",
    "ok_woman": "🙆‍♀️",
    "man-gesturing-ok": "🙆‍♂️",
    "woman-bowing": "🙇‍♀️",
    "man-bowing": "🙇‍♂️",
    "bow": "🙇‍♂️",
    "woman-raising-hand": "🙋‍♀️",
    "raising_hand": "🙋‍♀️",
    "man-raising-hand": "🙋‍♂️",
    "woman-frowning": "🙍‍♀️",
    "person_frowning": "🙍‍♀️",
    "man-frowning": "🙍‍♂️",
    "woman-pouting": "🙎‍♀️",
    "person_with_pouting_face": "🙎‍♀️",
    "man-pouting": "🙎‍♂️",
    "woman-rowing-boat": "🚣‍♀️",
    "man-rowing-boat": "🚣‍♂️",
    "rowboat": "🚣‍♂️",
    "woman-biking": "🚴‍♀️",
    "man-biking": "🚴‍♂️",
    "bicyclist": "🚴‍♂️",
    "woman-mountain-biking": "🚵‍♀️",
    "man-mountain-biking": "🚵‍♂️",
    "mountain_bicyclist": "🚵‍♂️",
    "woman-walking": "🚶‍♀️",
    "man-walking": "🚶‍♂️",
    "walking": "🚶‍♂️",
    "woman-facepalming": "🤦‍♀️",
    "man-facepalming": "🤦‍♂️",
    "woman-shrugging": "🤷‍♀️",
    "man-shrugging": "🤷‍♂️",
    "woman-cartwheeling": "🤸‍♀️",
    "man-cartwheeling": "🤸‍♂️",
    "woman-juggling": "🤹‍♀️",
    "man-juggling": "🤹‍♂️",
    "woman-wrestling": "🤼‍♀️",
    "man-wrestling": "🤼‍♂️",
    "woman-playing-water-polo": "🤽‍♀️",
    "man-playing-water-polo": "🤽‍♂️",
    "woman-playing-handball": "🤾‍♀️",
    "man-playing-handball": "🤾‍♂️",
    "woman-bouncing-ball": "⛹️‍♀️",
    "man-bouncing-ball": "⛹️‍♂️",
    "person_with_ball": "⛹️‍♂️",
}



"""A parallax gadget."""

from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Self

import numpy as np
from numpy.typing import NDArray

from ..colors import TRANSPARENT, AColor
from ..texture_tools import composite, read_texture, resize_texture
from .graphics import (
    Blitter,
    Graphics,
    Interpolation,
    Point,
    PosHint,
    Size,
    SizeHint,
    scale_geometry,
)

__all__ = ["Parallax", "Interpolation", "Point", "Size"]


class Parallax(Graphics):
    r"""
    A parallax gadget.

    Parameters
    ----------
    path : Path | None, default: None
        Path to directory of images for layers of the parallax (loaded in lexographical
        order of filenames) layered from background to foreground.
    speeds : Sequence[float] | None, default: None
        The scrolling speed of each layer. Default speeds are `1/(N - i)` where `N` is
        the number of layers and `i` is the index of a layer.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    layers : list[Image]
        Layers of the parallax.
    speeds : Sequence[float]
        The scrolling speed of each layer.
    offset : tuple[float, float]
        Vertical and horizontal offset of the parallax.
    vertical_offset : float
        Vertical offset of the parallax.
    horizontal_offset : float
        Horizontal offset of the parallax.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    from_textures(textures, ...)
        Create a :class:`Parallax` from an iterable of uint8 RGBA numpy array.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path | None = None,
        speeds: Sequence[float] | None = None,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.layers: list[NDArray[np.uint8]]
        """Layers of the parallax."""

        if path is None:
            self.layers = []
        else:
            paths = sorted(path.iterdir(), key=lambda file: file.name)
            self.layers = [read_texture(path) for path in paths]

        self.speeds: Sequence[float]
        """The scrolling speed of each layer."""

        nlayers = len(self.layers)
        if speeds is None:
            self.speeds = [1 / (nlayers - i) for i in range(nlayers)]
        elif nlayers != len(speeds):
            raise ValueError("number of layers doesn't match number of layer speeds")
        else:
            self.speeds = speeds

        self._vertical_offset = self._horizontal_offset = 0.0

        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

    def on_size(self):
        """Resize parallax layers."""
        h, w = scale_geometry(self._blitter, self._size)
        self.texture = np.empty((h, w, 4), np.uint8)
        self._sized_layers = [
            resize_texture(texture, (h, w), self.interpolation)
            for texture in self.layers
        ]
        self._update_texture()

    @property
    def vertical_offset(self) -> float:
        """Vertical offset of the parallax."""
        return self._vertical_offset

    @vertical_offset.setter
    def vertical_offset(self, offset: float):
        self._vertical_offset = offset
        self._update_texture()

    @property
    def horizontal_offset(self) -> float:
        """Horizontal offset of the parallax."""
        return self._horizontal_offset

    @horizontal_offset.setter
    def horizontal_offset(self, offset: float):
        self._horizontal_offset = offset
        self._update_texture()

    @property
    def offset(self) -> tuple[float, float]:
        """Vertical and horizontal offset of the parallax."""
        return self._vertical_offset, self._horizontal_offset

    @offset.setter
    def offset(self, offset: tuple[float, float]):
        self._vertical_offset, self._horizontal_offset = offset
        self._update_texture()

    def _update_texture(self):
        self.clear()
        for speed, texture in zip(self.speeds, self._sized_layers):
            rolls = (
                -round(speed * self._vertical_offset),
                -round(speed * self._horizontal_offset),
            )
            composite(np.roll(texture, rolls, axis=(0, 1)), self.texture)

    @classmethod
    def from_textures(
        cls,
        textures: Iterable[NDArray[np.uint8]],
        *,
        speeds: Sequence[float] | None = None,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Parallax` from an iterable of uint8 RGBA numpy array.

        Parameters
        ----------
        textures : Iterable[NDArray[np.uint8]]
            An iterable of RGBA textures that will be the layers of the parallax.
        speeds : Sequence[float] | None, default: None
            The scrolling speed of each layer. Default speeds are `1/(N - i)` where `N`
            is the number of layers and `i` is the index of a layer.
        default_color : AColor, default: AColor(0, 0, 0, 0)
            Default texture color.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        blitter : Blitter, default: "half"
            Determines how graphics are rendered.
        size : Size, default: Size(10, 10)
            Size of gadget.
        pos : Point, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Parallax
            A new parallax gadget.
        """
        layers = list(textures)
        nlayers = len(layers)
        if speeds is None:
            speeds = [1 / (nlayers - i) for i in range(nlayers)]
        elif nlayers != len(speeds):
            raise ValueError("number of layers doesn't match number of layer speeds")

        parallax = cls(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        parallax.layers = layers
        parallax.speeds = speeds
        parallax.on_size()
        return parallax



"""
A recycle-view provides a view into a large data-set using as few gadgets as
possible.
"""

from __future__ import annotations

from abc import abstractmethod

from ..geometry.regions import Region
from ..terminal.events import MouseButton
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .scroll_view import ScrollView

__all__ = ["RecycleView", "Point", "Size"]


class RecycleView[T, G: Gadget](ScrollView):
    r"""
    A recycle-view provides a view into a large data-set using as few gadgets as
    possible.

    ``RecycleView`` is an abstract class. The following methods require an
    implementation:

    * ``get_layout(i)``: Return the size and position of the ith item in
      ``recycle_view_data``.
    * ``new_data_view()``: Initialize and return a new data-view gadget for the recycle
      view.
    * ``update_data_view(data_view, datum)``: Update ``data_view`` to display ``datum``.

    In addition, one may want to re-implement ``set_view_size()`` which sets the size of
    the view of all the data.

    Parameters
    ----------
    recycle_view_data: list[T] | None, default: None
        The recycle-view's data.
    allow_vertical_scroll : bool, default: True
        Allow vertical scrolling.
    allow_horizontal_scroll : bool, default: True
        Allow horizontal scrolling.
    show_vertical_bar : bool, default: True
        Whether the vertical scrollbar is shown.
    show_horizontal_bar : bool, default: True
        Whether the horizontal scrollbar is shown.
    dynamic_bars : bool, default: False
        Whether bars are shown/hidden depending on the view size.
    scrollwheel_enabled : bool, default: True
        Allow vertical scrolling with scrollwheel.
    arrow_keys_enabled : bool, default: True
        Allow scrolling with arrow keys.
    inertial_scrolling_enabled : bool, default: True
        Whether inertial scrolling is enabled.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton, default: "left"
        Mouse button used for grabbing.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    recycle_view_data: list[T]
        The recycle-view's data.
    view : Gadget | None
        The scrolled gadget.
    allow_vertical_scroll : bool
        Allow vertical scrolling.
    allow_horizontal_scroll : bool
        Allow horizontal scrolling.
    show_vertical_bar : bool
        Whether the vertical scrollbar is shown.
    show_horizontal_bar : bool
        Whether the horizontal scrollbar is shown.
    dynamic_bars : bool, default: False
        Whether bars are shown/hidden depending on the view size.
    scrollwheel_enabled : bool
        Allow vertical scrolling with scrollwheel.
    arrow_keys_enabled : bool
        Allow scrolling with arrow keys.
    inertial_scrolling_enabled : bool
        Whether inertial scrolling is enabled.
    vertical_proportion : float
        Vertical scroll position as a proportion of total height.
    horizontal_proportion : float
        Horizontal scroll position as a proportion of total width.
    port_height : int
        Height of view port.
    port_width : int
        Width of view port.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent : Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    get_layout(i)
        Return the size and position of the ith item of the recycle-view's data.
    new_data_view()
        Initialize and return a new data-view gadget for the recycle-view.
    update_data_view(data_view, datum)
        Update ``data_view`` to display ``datum``.
    set_view_size()
        Set the size of the view.
    refresh_data()
        Refresh data of currently visible data-views.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        recycle_view_data: list[T] | None = None,
        allow_vertical_scroll: bool = True,
        allow_horizontal_scroll: bool = True,
        show_vertical_bar: bool = True,
        show_horizontal_bar: bool = True,
        dynamic_bars: bool = False,
        scrollwheel_enabled: bool = True,
        arrow_keys_enabled: bool = True,
        inertial_scrolling_enabled: bool = True,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton = "left",
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            allow_vertical_scroll=allow_vertical_scroll,
            allow_horizontal_scroll=allow_horizontal_scroll,
            show_vertical_bar=show_vertical_bar,
            show_horizontal_bar=show_horizontal_bar,
            dynamic_bars=dynamic_bars,
            scrollwheel_enabled=scrollwheel_enabled,
            arrow_keys_enabled=arrow_keys_enabled,
            inertial_scrolling_enabled=inertial_scrolling_enabled,
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._unused_views: list[G] = []
        """Currently unused views."""
        self._index_to_view: dict[int, G] = {}
        """Index of datum to the view displaying it."""
        self._view_pos_bind: int | None = None
        """uid to unbind view's ``pos`` from ``_update_view``."""
        self._old_len_of_data: int = 0
        """
        Previous length of recycle-view data used to determine whether layout should be
        recalculated in ``refresh_data``.
        """

        self.recycle_view_data: list[T]
        """The recycle-view's data."""
        if recycle_view_data is None:
            self.recycle_view_data = []
        else:
            self.recycle_view_data = recycle_view_data

        self.view = Gadget()
        self.set_view_size()
        self.refresh_data()

    @abstractmethod
    def get_layout(self, i: int) -> tuple[Size, Point]:
        """
        Return the size and position of the ith item of the recycle-view's data.

        Parameters
        ----------
        i : int
            The index of an item in the recycle-view's data.

        Returns
        -------
        tuple[Size, Point]
            The size and position of the item.
        """

    @abstractmethod
    def new_data_view(self) -> G:
        """Initialize and return a new data-view gadget for the recycle-view."""

    @abstractmethod
    def update_data_view(self, data_view: G, datum: T) -> None:
        """Update ``data_view`` to display ``datum``."""

    def set_view_size(self) -> None:
        """
        Set the size of the view.

        The default implementation sets the view size to be the minimum size required to
        fully encompass the bottom-most, right-most rect returned by ``get_layout``.
        """
        if self.view is None:
            return

        bottom = right = 0
        for i in range(len(self.recycle_view_data)):
            size, pos = self.get_layout(i)
            current_bottom = size.height + pos.y
            if current_bottom > bottom:
                bottom = current_bottom

            current_right = size.width + pos.x
            if current_right > right:
                right = current_right

        self.view.size = Size(bottom, right)

    def refresh_data(self) -> None:
        """Refresh data of currently visible data-views."""
        if self._old_len_of_data == len(self.recycle_view_data):
            for i, data_view in self._index_to_view.items():
                self.update_data_view(data_view, self.recycle_view_data[i])
        else:
            self._old_len_of_data = len(self.recycle_view_data)
            for data_view in self._index_to_view.values():
                self.view.remove_gadget(data_view)
            self._unused_views.extend(self._index_to_view.values())
            self._index_to_view.clear()
            self.set_view_size()
            self._update_recycle_view()

    def on_add(self) -> None:
        """Bind the view's size and pos to update the items in the view."""
        super().on_add()
        self._view_pos_bind = self.view.bind("pos", self._update_recycle_view)

    def on_remove(self) -> None:
        """Unbind the view's size and pos from updating the items in the view."""
        if self.view is not None and self._view_pos_bind is not None:
            self.view.unbind(self._view_pos_bind)
        self._view_pos_bind = None
        super().on_remove()

    def on_size(self) -> None:
        """Update the recycle-view on resize."""
        super().on_size()
        if self.view is not None:
            self._update_recycle_view()

    def _update_recycle_view(self) -> None:
        scroll_view_reg = Region.from_rect(
            -self.view.pos, (self.port_height, self.port_width)
        )
        recycle_view_reg = Region.from_rect((0, 0), self.view.size)
        clipping_rect = scroll_view_reg & recycle_view_reg

        seen: list[int] = []
        for i in range(len(self.recycle_view_data)):
            size, pos = self.get_layout(i)
            datum_reg = Region.from_rect(pos, size)
            is_seen = bool(datum_reg & clipping_rect)
            has_gadget = i in self._index_to_view
            if has_gadget != is_seen:
                if has_gadget:
                    gadget = self._index_to_view[i]
                    self.view.remove_gadget(gadget)
                    self._unused_views.append(gadget)
                    del self._index_to_view[i]
                else:
                    seen.append(i)

        if len(seen) > len(self._unused_views):
            new_items = len(seen) - len(self._unused_views)
            self._unused_views.extend(self.new_data_view() for _ in range(new_items))

        for i in seen:
            gadget = self._unused_views.pop()
            self._index_to_view[i] = gadget
            gadget.size, gadget.pos = self.get_layout(i)
            self.update_data_view(gadget, self.recycle_view_data[i])
            self.view.add_gadget(gadget)



"""Functions for blending colors and creating color gradients."""

from itertools import pairwise
from math import sin, tau

from ..geometry import lerp
from ..geometry.easings import EASINGS, Easing
from .color_types import AColor, Color

__all__ = ["darken_only", "gradient", "lerp_colors", "lighten_only", "rainbow_gradient"]


def darken_only(a: Color, b: Color) -> Color:
    """
    Return a color that is the minimum of each channel in ``a`` and ``b``.

    Parameters
    ----------
    a : Color
        A color.
    b : Color
        A color.

    Returns
    -------
    Color
        A color with smallest components of ``a`` and ``b``.
    """
    color = (min(c1, c2) for c1, c2 in zip(a, b))
    return Color(*color)


def lighten_only(a: Color, b: Color) -> Color:
    """
    Return a color that is the maximum of each channel in ``a`` and ``b``.

    Parameters
    ----------
    a : Color
        A color.
    b : Color
        A color.

    Returns
    -------
    Color
        A color with largest components of ``a`` and ``b``.
    """
    color = (max(c1, c2) for c1, c2 in zip(a, b))
    return Color(*color)


def lerp_colors[T: (Color, AColor, tuple[int, ...])](a: T, b: T, p: float) -> T:
    """
    Linear interpolation from ``a`` to ``b`` with proportion ``p``.

    Parameters
    ----------
    a : (Color, AColor, tuple[int, ...])
        A color.
    b : (Color, AColor, tuple[int, ...])
        A color.
    p : float
        Proportion from a to b.

    Returns
    -------
    (Color, AColor, tuple[int, ...])
        The linear interpolation of ``a`` and ``b``.
    """
    color = (round(lerp(c1, c2, p)) for c1, c2 in zip(a, b))
    if isinstance(a, (Color, AColor)):
        return type(a)(*color)
    return tuple(color)


def gradient[T: (Color, AColor, tuple[int, ...])](
    *color_stops: T, n: int, easing: Easing = "linear"
) -> list[T]:
    r"""
    Return a smooth gradient of length ``n`` between all colors in ``color_stops``.

    Parameters
    ----------
    \*color_stops : (Color, AColor, tuple[int, ...])
        Colors between each gradient.
    n : int
        Length of gradient. Must be equal to or larger than ``color_stops``.
    easing : Easing, default: "linear"
        Easing applied to interpolations between ``color stops``.

    Returns
    -------
    list[(Color, AColor, tuple[int, ...])]
        A smooth gradient between all colors in ``color_stops``.
    """
    ncolors = len(color_stops)
    if n < ncolors:
        raise ValueError(f"gradient too small to contain all color stops ({n=})")
    if ncolors == 1:
        return [color_stops[0]] * n

    ease = EASINGS[easing]
    d, r = divmod(n - ncolors, ncolors - 1)
    gradient = []
    for i, (a, b) in enumerate(pairwise(color_stops)):
        gradient.append(a)
        len_ = d + (i < r)
        gradient.extend(
            lerp_colors(a, b, ease((j + 1) / (len_ + 1))) for j in range(len_)
        )
    gradient.append(b)
    return gradient


def rainbow_gradient(n: int, *, alpha: int | None = None) -> list[Color | AColor]:
    """
    Return a rainbow gradient of ``n`` colors.

    Parameters
    ----------
    n : int
        Number of colors in gradient.
    alpha : int | None, default: None
        If ``alpha`` is not given, gradient colors will have no alpha channel.
        Otherwise, the color's alpha channel is given by ``alpha``.

    Returns
    -------
    list[Color | AColor]
        A rainbow gradient of colors.
    """
    theta = tau / n
    offsets = [0, tau / 3, 2 * tau / 3]

    def color(i):
        return (int(sin(i * theta + offset) * 127 + 128) for offset in offsets)

    if alpha is None:
        return [Color(*color(i)) for i in range(n)]

    return [AColor(*color(i), alpha) for i in range(n)]



"""A raycaster gadget."""

import numpy as np
from numpy.typing import NDArray

from ..colors import ABLACK, TRANSPARENT, AColor
from ._raycasting import cast_rays
from .graphics import Blitter, Graphics, Interpolation, Point, PosHint, Size, SizeHint

__all__ = ["Raycaster", "Point", "Size"]


class Raycaster(Graphics):
    r"""
    A raycaster gadget.

    ``caster_map`` should not contain a value greater than the length of
    ``wall_textures``. A non-zero value ``N`` in ``caster_map`` represents a wall with
    texture ``wall_textures[N - 1]``.

    ``sprite_indexes`` should not contain a value greater than or equal to the length of
    ``sprite_textures``. A value ``N`` in ``sprite_indexes`` represents a sprite with
    texture ``sprite_textures[N]``. ``sprite_coords`` and ``sprite_indexes`` must be the
    same length.

    Parameters
    ----------
    caster_map : NDArray[np.uint8]
        The raycaster map.
    wall_textures : List[NDArray[np.uint8]]
        Textures for walls.
    camera_coord : tuple[float, float], default: (0.0, 0.0)
        The camera's position.
    camera_angle : float, default: 0.0
        The camera's angle.
    camera_fov : float, default: 0.66
        The camera's field-of-view. Somewhere between 0-1.
    sprite_coords : NDArray[np.float64] | None, default: None
        Positions of sprites.
    sprite_indexes : NDArray[np.uint8] | None, default: None
        Texture indexes of sprites.
    sprite_textures : list[NDArray[np.uint8]] | None, default: None
        Textures for sprites.
    ceiling : NDArray[np.uint8] | None, default: None
        Optional ceiling texture.
    ceiling_color : AColor, default: ABLACK
        Color of ceiling if no ceiling texture.
    floor : NDArray[np.uint8] | None, default: None
        Optional floor texture.
    floor_color : AColor, default: ABLACK
        Color of floor if no floor texture.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    caster_map : NDArray[np.uint8]
        The raycaster map.
    wall_textures : List[NDArray[np.uint8]]
        Textures for walls.
    camera_coord : tuple[float, float]
        The camera's position.
    camera_angle : float
        The camera's angle.
    camera_fov : float
        The camera's field-of-view. Somewhere between 0-1.
    sprite_coords : NDArray[np.float64]
        Positions of sprites.
    sprite_indexes : NDArray[np.uint8]
        Texture indexes of sprites.
    sprite_textures : list[NDArray[np.uint8]]
        Textures for sprites.
    ceiling : NDArray[np.uint8] | None
        The ceiling texture.
    ceiling_color : AColor
        Color of ceiling if no ceiling texture.
    floor : NDArray[np.uint8] | None
        The floor texture.
    floor_color : AColor
        Color of floor if no floor texture.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    cast_rays()
        Update texture by casting all rays.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        caster_map: NDArray[np.uint8],
        wall_textures: list[NDArray[np.uint8]],
        camera_coord: tuple[float, float] = (0.0, 0.0),
        camera_angle: float = 0.0,
        camera_fov: float = 0.66,
        sprite_coords: NDArray[np.float64] | None = None,
        sprite_indexes: NDArray[np.uint8] | None = None,
        sprite_textures: list[NDArray[np.uint8]] | None = None,
        ceiling: NDArray[np.uint8] | None = None,
        ceiling_color: AColor = ABLACK,
        floor: NDArray[np.uint8] | None = None,
        floor_color: AColor = ABLACK,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.caster_map = caster_map
        """The raycaster map."""
        self.wall_textures = wall_textures
        """Textures for walls."""
        self.camera_coord = camera_coord
        """The camera's position."""
        self.camera_angle = camera_angle
        """The camera's angle."""
        self.camera_fov = camera_fov
        """The camera's field-of-view. Somewhere between 0-1."""
        self.sprite_coords: NDArray[np.float64]
        """Positions of sprites."""
        if sprite_coords is None:
            self.sprite_coords = np.empty((0, 2), np.float64)
        else:
            self.sprite_coords = sprite_coords

        self.sprite_indexes: NDArray[np.uint8]
        """Texture indexes of sprites."""
        if sprite_indexes is None:
            self.sprite_indexes = np.empty(0, np.uint8)
        else:
            self.sprite_indexes = sprite_indexes

        self.sprite_textures = sprite_textures
        """Textures for sprites."""
        self.ceiling = ceiling
        """Optional ceiling texture."""
        self.ceiling_color = ceiling_color
        """Color of ceiling if no ceiling texture."""
        self.floor = floor
        """Optional floor texture."""
        self.floor_color = floor_color
        """Color of floor if no floor texture."""

        self.on_size()

    def cast_rays(self):
        """Update texture by casting all rays."""
        h = self.texture.shape[0]
        self.texture[: h // 2] = self.ceiling_color
        self.texture[h // 2 :] = self.floor_color
        cast_rays(
            self.texture,
            self.caster_map,
            self.camera_coord,
            self.camera_angle,
            self.camera_fov,
            self.wall_textures,
            self.ceiling,
            self.floor,
            self.sprite_indexes,
            self.sprite_coords,
            self.sprite_textures,
        )
